# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools
import io
import os
import typing

from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack import ironic
from tobiko.openstack import metalsmith
from tobiko.openstack import neutron
from tobiko.openstack import topology
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.tripleo import _undercloud


CONF = config.CONF
LOG = log.getLogger(__name__)
OVN_RAFT = 'RAFT'
OVN_HA = 'HA'


def load_overcloud_rcfile() -> typing.Dict[str, str]:
    conf = tobiko.tobiko_config().tripleo
    return _undercloud.fetch_os_env(*conf.overcloud_rcfile)


@functools.lru_cache()
def has_overcloud(min_version: str = None,
                  max_version: str = None) -> bool:
    try:
        check_overcloud(min_version=min_version,
                        max_version=max_version)
    except (OvercloudNotFound, OvercloudVersionMismatch) as ex:
        LOG.debug(f'Overcloud not found: {ex}')
        return False
    else:
        LOG.debug('Overcloud found')
        return True


skip_if_missing_overcloud = tobiko.skip_unless(
    'TripleO overcloud not configured', has_overcloud)


def skip_unless_has_overcloud(min_version: tobiko.VersionType = None,
                              max_version: tobiko.VersionType = None):
    return tobiko.skip_on_error(
        'TripleO overcloud not configured',
        check_overcloud,
        min_version=min_version,
        max_version=max_version,
        error_type=(OvercloudNotFound, OvercloudVersionMismatch))


class OvercloudKeystoneCredentialsFixtureBase(
        _undercloud.UndercloudKeystoneCredentialsFixtureBase):

    def _get_environ(self) -> typing.Dict[str, str]:
        return load_overcloud_rcfile()


class OvercloudKeystoneCredentialsFixture(
        OvercloudKeystoneCredentialsFixtureBase,
        keystone.DelegateKeystoneCredentialsFixture):

    @staticmethod
    def _get_delegates() -> typing.List[keystone.KeystoneCredentialsFixture]:
        return [
            tobiko.get_fixture(
                OvercloudCloudsFileKeystoneCredentialsFixture),
            tobiko.get_fixture(
                OvercloudEnvironKeystoneCredentialsFixture)]


def list_overcloud_nodes(**params):
    session = _undercloud.undercloud_keystone_session()
    client = metalsmith.get_metalsmith_client(session=session)
    return metalsmith.list_instances(client=client, **params)


def find_overcloud_node(**params):
    session = _undercloud.undercloud_keystone_session()
    client = metalsmith.get_metalsmith_client(session=session)
    return metalsmith.find_instance(client=client, **params)


def power_on_overcloud_node(instance: metalsmith.MetalsmithInstance,
                            timeout: tobiko.Seconds = 120.,
                            sleep_time: tobiko.Seconds = 5.):
    session = _undercloud.undercloud_keystone_session()
    client = ironic.get_ironic_client(session=session)
    ironic.power_on_node(client=client,
                         node=instance.uuid,
                         timeout=timeout,
                         sleep_time=sleep_time)


def power_off_overcloud_node(instance: metalsmith.MetalsmithInstance,
                             timeout: tobiko.Seconds = None,
                             sleep_time: tobiko.Seconds = None):
    session = _undercloud.undercloud_keystone_session()
    client = ironic.get_ironic_client(session=session)
    ironic.power_off_node(client=client,
                          node=instance.uuid,
                          timeout=timeout,
                          sleep_time=sleep_time)


def overcloud_ssh_client(ip_version: int = None,
                         network_name: str = None,
                         instance: metalsmith.MetalsmithInstance = None,
                         host_config=None):
    if host_config is None:
        host_config = overcloud_host_config(ip_version=ip_version,
                                            network_name=network_name,
                                            instance=instance)
    tobiko.check_valid_type(host_config.host, str)
    return ssh.ssh_client(host=host_config.host,
                          **host_config.connect_parameters)


def overcloud_host_config(ip_version: int = None,
                          network_name: str = None,
                          instance: metalsmith.MetalsmithInstance = None):
    host_config = OvercloudHostConfig(ip_version=ip_version,
                                      network_name=network_name,
                                      instance=instance)
    return tobiko.setup_fixture(host_config)


def overcloud_node_ip_address(ip_version: int = None,
                              network_name: str = None,
                              instance: metalsmith.MetalsmithInstance = None,
                              **params):
    if instance is None:
        instance = find_overcloud_node(**params)
    ip_version = ip_version or CONF.tobiko.tripleo.overcloud_ip_version
    network_name = network_name or CONF.tobiko.tripleo.overcloud_network_name
    address = metalsmith.find_instance_ip_address(instance=instance,
                                                  ip_version=ip_version,
                                                  network_name=network_name)
    LOG.debug(f"Got Overcloud node address '{address}' from Undercloud "
              f"(ip_version={ip_version}, network_name={network_name}, "
              f"instance={instance})")
    return address


class OvercloudSshKeyFileFixture(tobiko.SharedFixture):

    @property
    def key_filename(self):
        return tobiko.tobiko_config_path(
            CONF.tobiko.tripleo.overcloud_ssh_key_filename)

    def setup_fixture(self):
        self.setup_key_file()

    def setup_key_file(self):
        key_filename = self.key_filename
        key_dirname = os.path.dirname(key_filename)
        tobiko.makedirs(key_dirname, mode=0o700)

        ssh_client = _undercloud.undercloud_ssh_client()
        _get_undercloud_file(ssh_client=ssh_client,
                             source='~/.ssh/id_rsa',
                             destination=key_filename,
                             mode=0o600)
        _get_undercloud_file(ssh_client=ssh_client,
                             source='~/.ssh/id_rsa.pub',
                             destination=key_filename + '.pub',
                             mode=0o600)


def _get_undercloud_file(ssh_client, source, destination, mode):
    content = sh.execute(['cat', source],
                         ssh_client=ssh_client).stdout
    with io.open(destination, 'wb') as fd:
        fd.write(content.encode())
    os.chmod(destination, mode)


class OvercloudHostConfig(tobiko.SharedFixture):

    key_file = tobiko.required_fixture(OvercloudSshKeyFileFixture)

    def __init__(self,
                 host: str = None,
                 hostname: str = None,
                 ip_version: int = None,
                 instance: metalsmith.MetalsmithInstance = None,
                 key_filename: str = None,
                 network_name: str = None,
                 port: int = None,
                 username: str = None,
                 **kwargs):
        super(OvercloudHostConfig, self).__init__()
        self.host = host
        self.instance = instance
        self.ip_version = ip_version
        self.key_filename = key_filename
        self.host = host
        self.hostname = hostname
        self.network_name = network_name
        self.port = port
        self.username = username
        self._connect_parameters = ssh.gather_ssh_connect_parameters(**kwargs)

    def setup_fixture(self):
        if self.hostname is None:
            self.hostname = str(overcloud_node_ip_address(
                hostname=self.host,
                ip_version=self.ip_version,
                network_name=self.network_name,
                instance=self.instance))
        if self.host is None:
            self.host = self.hostname
        if self.port is None:
            self.port = CONF.tobiko.tripleo.overcloud_ssh_port
        if self.username is None:
            self.username = CONF.tobiko.tripleo.overcloud_ssh_username
        if self.key_filename is None:
            self.key_filename = self.key_file.key_filename

    @property
    def connect_parameters(self):
        parameters = ssh.ssh_host_config(
            host=str(self.hostname)).connect_parameters
        parameters.update(ssh.gather_ssh_connect_parameters(self))
        parameters.update(self._connect_parameters)
        return parameters


def get_overcloud_nodes_dataframe(
        oc_node_df_function: typing.Callable[[ssh.SSHClientType],
                                             typing.Any]):
    """
     :param oc_node_df_function : a function that queries a oc node
     using a cli command and returns a datraframe with an added
     hostname field.

     This function concats oc nodes dataframes into a unified overcloud
     dataframe, seperated by hostname field

    :return: dataframe of all overcloud nodes processes
    """
    import pandas
    oc_nodes_dfs = list()
    for instance in list_overcloud_nodes():
        ssh_client = overcloud_ssh_client(instance=instance)
        oc_nodes_dfs.append(oc_node_df_function(ssh_client))
    oc_procs_df = pandas.concat(oc_nodes_dfs, ignore_index=True)
    return oc_procs_df


def is_redis_expected():
    if topology.verify_osp_version('17.0', lower=True):
        return True
    services_requiring_redis = (
        'designate', 'octavia', 'ceilometer', 'gnocchi', 'panko')
    for service in services_requiring_redis:
        if keystone.has_service(name=service):
            return True
    return False


class InvalidDBServiceModel(tobiko.TobikoException):
    pass


@functools.lru_cache()
def get_ovn_db_service_model():
    """Show in which mode OVN databases are configured

    There are two modes currently supported:
     - RAFT aka clustered service model (default starting OSP17.0)
     - HA aka Active-Backup service model (default for pre-OSP17.0 versions)

    For more information:
    https://docs.openvswitch.org/en/latest/ref/ovsdb.7/#service-models
    """
    if not neutron.has_ovn():
        raise InvalidDBServiceModel('OVN is not enabled')
    controller0 = topology.list_openstack_nodes(group='controller')[0]
    db_info = sh.execute('find / -name ovnnb_db.db | xargs sudo head -n 1',
                         ssh_client=controller0.ssh_client, sudo=True)
    if 'CLUSTER' in db_info.stdout:
        return OVN_RAFT
    elif 'JSON' in db_info.stdout:
        return OVN_HA
    else:
        msg = 'Unable to fetch OVN DB service model from {} string. Only RAFT'\
              ' and HA models are supported'.format(db_info.stdout)
        LOG.error(msg)
        raise InvalidDBServiceModel(message=msg)


@functools.lru_cache()
def is_ovn_using_raft():
    try:
        if get_ovn_db_service_model() == OVN_RAFT:
            return True
    except InvalidDBServiceModel:
        pass
    return False


@functools.lru_cache()
def is_ovn_using_ha():
    try:
        if get_ovn_db_service_model() == OVN_HA:
            return True
    except InvalidDBServiceModel:
        pass
    return False


skip_unless_ovn_using_raft = tobiko.skip_unless(
    'OVN does not use RAFT DB model', is_ovn_using_raft)


skip_unless_ovn_using_ha = tobiko.skip_unless(
    'OVN does not use HA DB model', is_ovn_using_ha)


@functools.lru_cache()
def are_kexec_tools_installed():
    for controller in topology.list_openstack_nodes(group='controller'):
        try:
            sh.execute('rpm -q kexec-tools', ssh_client=controller.ssh_client)
        except sh.ShellCommandFailed:
            return False
    return True


skip_unless_kexec_tools_installed = tobiko.skip_unless(
    'kexec-tools package not installed on controller nodes',
    are_kexec_tools_installed)


def overcloud_version() -> tobiko.Version:
    from tobiko.tripleo import _topology
    node = topology.find_openstack_node(group='overcloud')
    assert isinstance(node, _topology.TripleoTopologyNode)
    return node.rhosp_version


@functools.lru_cache()
def check_overcloud(min_version: str = None,
                    max_version: str = None):
    try:
        _undercloud.check_undercloud()
    except _undercloud.UndercloudNotFound as ex:
        raise OvercloudNotFound(cause=f'undercloud not found: {ex}') from ex

    if min_version or max_version:
        tobiko.check_version(overcloud_version(),
                             min_version=min_version,
                             max_version=max_version,
                             mismatch_error=OvercloudVersionMismatch)


class OvercloudNotFound(tobiko.ObjectNotFound):
    message = 'overcloud not found: {cause}'


class OvercloudVersionMismatch(tobiko.VersionMismatch):
    message = 'overcloud version mismatch: {version} {cause}'


class OvercloudCloudsFileKeystoneCredentialsFixture(
        OvercloudKeystoneCredentialsFixtureBase,
        keystone.CloudsFileKeystoneCredentialsFixture):

    @staticmethod
    def _get_default_cloud_name() -> typing.Optional[str]:
        return tobiko.tobiko_config().tripleo.overcloud_cloud_name


class OvercloudEnvironKeystoneCredentialsFixture(
        OvercloudKeystoneCredentialsFixtureBase,
        keystone.EnvironKeystoneCredentialsFixture):
    pass


def overcloud_keystone_session() -> keystone.KeystoneSession:
    credentials = overcloud_keystone_credentials()
    return keystone.get_keystone_session(credentials=credentials)


def overcloud_keystone_credentials() -> keystone.KeystoneCredentialsFixture:
    return tobiko.get_fixture(OvercloudKeystoneCredentialsFixture)


def overcloud_keystone_client() -> keystone.KeystoneClient:
    session = overcloud_keystone_session()
    return keystone.get_keystone_client(session=session)


def setup_overcloud_keystone_credentials():
    if has_overcloud():
        keystone.register_default_keystone_credentials(
            credentials=overcloud_keystone_credentials(),
            position=0)



# Copyright (c) 2021 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import pytest
import testtools

import tobiko
from tobiko import config
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.shell import ip


CONF = config.CONF
LOG = log.getLogger(__name__)


class RebootTrunkServerStackFixture(stacks.UbuntuServerStackFixture):

    def validate_created_stack(self):
        # (fressi) must wait cloud init to complete Nova
        # server setup before shutting it down so that we ensure all
        # network devices gets persistent after reboots
        stack = super().validate_created_stack()
        nova.activate_server(server=self.server_id)
        self.wait_for_cloud_init_done()
        return stack


@neutron.skip_if_missing_networking_extensions('trunk')
class RebootTrunkTest(testtools.TestCase):
    """Tests trunk functionality"""

    stack = tobiko.required_fixture(RebootTrunkServerStackFixture)

    def test_0_vlan_ip_addresses(self):
        """Check Nova server VLAN port IP addresses"""
        self.stack.ensure_server_status('ACTIVE')
        expected = set(self.stack.list_vlan_fixed_ips())
        for attempt in tobiko.retry(timeout=300, interval=10):
            actual = set(ip.list_ip_addresses(device=self.stack.vlan_device,
                                              ssh_client=self.stack.ssh_client,
                                              scope='global'))
            unexpected = actual - expected
            if unexpected:
                self.fail("Unexpected IP address assigned to VLAN port: "
                          f"{unexpected}")

            missing = expected - actual
            if missing:
                if attempt.is_last:
                    self.fail("IP addresses not assigned to VLAN port: "
                              f"{unexpected}")
                else:
                    LOG.debug("IP addresses still not assigned to VLAN port: "
                              f"{unexpected}")
            else:
                break
        else:
            raise RuntimeError("Broken retry loop")
        self.assertEqual(set(expected), set(actual))

    def test_1_ping_vlan_port(self):
        """Check Nova server VLAN port is reachable"""
        self.stack.ensure_server_status('ACTIVE')
        self.stack.assert_vlan_is_reachable()

    @pytest.mark.ovn_migration
    def test_2_ping_vlan_port_after_restart(self):
        """Check Nova server VLAN port is reachable after hard restart"""
        self.stack.ensure_server_status('SHUTOFF')
        self.stack.assert_vlan_is_unreachable()

        self.stack.ensure_server_status('ACTIVE')
        self.stack.assert_vlan_is_reachable()



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json
import re
import typing

import netaddr
import pytest
from neutron_lib import constants
from oslo_log import log
import testtools

import tobiko
from tobiko.shell import ping
from tobiko.shell import ip
from tobiko.shell import sh
from tobiko.openstack import neutron
from tobiko.openstack import stacks
from tobiko.openstack import topology


LOG = log.getLogger(__name__)
IPV4 = constants.IP_VERSION_4
IPV6 = constants.IP_VERSION_6


@pytest.mark.minimal
class PortTest(testtools.TestCase):
    """Test Neutron ports"""

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(stacks.CirrosServerStackFixture)

    def test_port_ips(self, ip_version: typing.Optional[int] = None):
        """Checks port IPS has been assigned to server via DHCP protocol"""
        port_ips = set(neutron.list_device_ip_addresses(
            device=self.stack.server_id,
            network=self.stack.network_stack.network_id,
            need_dhcp=self.stack.need_dhcp,
            ip_version=ip_version))
        if port_ips:
            # verify neutron port IPs and VM port IPs match
            # when a VM connected to the external network has been just
            # created, it may need some time to receive its IPv6 address
            for attempt in tobiko.retry(timeout=60., interval=4.):
                server_ips = set(ip.list_ip_addresses(
                    scope='global', ssh_client=self.stack.ssh_client))
                server_ips &= port_ips  # ignore other server IPs
                LOG.debug("Neutron IPs and VM IPs should match...")
                try:
                    self.assertEqual(
                        port_ips, server_ips,
                        f"Server {self.stack.server_id} is missing port "
                        f"IP(s): {port_ips - server_ips}")
                    break
                except self.failureException:
                    attempt.check_limits()
        elif ip_version:
            self.skipTest(f"Server has any port IPv{ip_version} address to be"
                          " tested")
        else:
            self.skipTest("Server has any port IP address to be tested")

    def test_port_network(self):
        self.assertEqual(self.stack.network_stack.network_id,
                         self.stack.port_details['network_id'])

    def test_port_subnets(self):
        """Checks port subnets"""
        port_subnets = [fixed_ip['subnet_id']
                        for fixed_ip in self.stack.port_details['fixed_ips']]
        network_subnets = self.stack.network_stack.network_details['subnets']
        self.assertEqual(set(network_subnets), set(port_subnets))

    def test_ping_subnet_gateways(self):
        """Checks server can ping its gateway IPs"""
        network_id = self.stack.network_stack.network_id
        subnets = neutron.list_subnets(network_id=network_id)
        LOG.debug("Subnets enabled are:\n"
                  f"{json.dumps(subnets, indent=4, sort_keys=True)}")
        gateway_ips = [netaddr.IPAddress(subnet['gateway_ip'])
                       for subnet in subnets]
        LOG.debug(f"Gateway IPs are: {gateway_ips}")
        ping.assert_reachable_hosts(gateway_ips,
                                    ssh_client=self.stack.ssh_client)

    def test_ping_port(self, network_id=None, device_id=None, ip_version=None):
        """Checks server can ping its own port"""
        port_ips = neutron.list_device_ip_addresses(
            device=device_id or self.stack.server_id,
            network=network_id or self.stack.network_stack.network_id,
            need_dhcp=self.stack.need_dhcp,
            ip_version=ip_version)
        if port_ips:
            ping.assert_reachable_hosts(port_ips,
                                        timeout=600.,
                                        ssh_client=self.stack.peer_ssh_client)
        elif ip_version:
            self.skipTest(f"Server has any port IPv{ip_version} address to be"
                          " tested")
        else:
            self.skipTest("Server has any port IP address to be tested")


# --- Test opening ports on external network ----------------------------------

@stacks.skip_unless_has_external_network
class UbuntuExternalPortTest(PortTest):
    """Test Neutron ports"""

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(
        stacks.UbuntuExternalServerStackFixture)


# --- Test la-h3 extension ----------------------------------------------------

@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class L3HAPortTest(PortTest):
    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_fixture(stacks.L3haServerStackFixture)


class ExtraDhcpOptsPortTest(PortTest):
    """Test extra-dhcp-options port parameter"""
    stack = tobiko.required_fixture(
        stacks.ExtraDhcpOptsCirrosServerStackFixture)

    def test_extra_dhcp_opts(self):
        port = neutron.get_port(self.stack.port_id)
        for option in port['extra_dhcp_opts']:
            if 'domain-name' == option['opt_name']:
                domain = option['opt_value'].replace('"', '')
                break
        else:
            tobiko.fail('No extra-dhcp-opt found for domain-name')

        vm_resolv_conf = sh.execute('cat /etc/resolv.conf',
                                    ssh_client=self.stack.ssh_client).stdout
        self.assertIsNotNone(
            re.search(r'^search\s+{domain}$'.format(domain=domain),
                      vm_resolv_conf,
                      re.MULTILINE))


@neutron.skip_unless_is_ovn()
class ExtraDhcpOptsPortLoggingTest(testtools.TestCase):

    stack = tobiko.required_fixture(stacks.NetworkStackFixture)

    def test_extra_dhcp_opts_logs_unsupported_options(self):
        # initialize logs that match the pattern
        topology.assert_ovn_unsupported_dhcp_option_messages()

        wrong_ipv4_option = 'wrong-ipv4-option'
        wrong_ipv6_option = 'bananas'
        a_valid_ipv4_option_used_for_ipv6 = 'log-server'
        extra_dhcp_opts = [
            {'opt_value': '1.1.1.1',
             'opt_name': a_valid_ipv4_option_used_for_ipv6,
             'ip_version': IPV6},
            {'opt_value': 'ipv6.domain',
             'opt_name': 'domain-search',
             'ip_version': IPV6},
            {'opt_value': '1600',
             'opt_name': 'mtu',
             'ip_version': IPV4},
            {'opt_value': 'blablabla',
             'opt_name': wrong_ipv4_option,
             'ip_version': IPV4}]
        # create port with extra-dhcp-opts
        port = neutron.create_port(**{'network_id': self.stack.network_id,
                                      'extra_dhcp_opts': extra_dhcp_opts})
        self.addCleanup(neutron.delete_port, port['id'])
        # find new logs that match the pattern
        invalid_options = [wrong_ipv4_option,
                           a_valid_ipv4_option_used_for_ipv6]
        # assert every invalid dhcp option is logged
        topology.assert_ovn_unsupported_dhcp_option_messages(
            unsupported_options=invalid_options,
            port_uuid=port['id'])

        extra_dhcp_opts.append({'opt_value': '1.1.1.1',
                                'opt_name': wrong_ipv6_option,
                                'ip_version': IPV6})
        # update port with new extra-dhcp-opts
        port = neutron.update_port(port['id'],
                                   **{'extra_dhcp_opts': extra_dhcp_opts})
        invalid_options.append(wrong_ipv6_option)
        # assert every invalid dhcp option is logged
        topology.assert_ovn_unsupported_dhcp_option_messages(
            unsupported_options=invalid_options,
            port_uuid=port['id'])



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import contextlib
import os
import sys
import typing
import types
import unittest

from oslo_log import log
import testtools

import tobiko
from tobiko.common import _exception
from tobiko.common import _retry
from tobiko.common import _time


LOG = log.getLogger(__name__)

os.environ.setdefault('PYTHON', sys.executable)

TestResult = unittest.TestResult
TestCase = unittest.TestCase
TestSuite = unittest.TestSuite


class BaseTestCase(unittest.TestCase):
    """Base test case class for tobiko test cases

    The reason this for exist is to have a way to override other tools base
    classes methods
    """

    _subtest: typing.Optional[unittest.TestCase] = None

    def run(self, result: TestResult = None) -> typing.Optional[TestResult]:
        with enter_test_case(self):
            return super().run(result)


class TestToolsTestCase(BaseTestCase, testtools.TestCase):
    pass


class TestCaseEntry(typing.NamedTuple):
    case: TestCase
    start_time: float


class DummyTestCase(BaseTestCase):

    def runTest(self):
        raise RuntimeError('Dummy test case')

    @contextlib.contextmanager
    def subTest(self, msg: typing.Any = ..., **params) \
            -> typing.Iterator[None]:
        yield


class TestCaseManager:

    def __init__(self,
                 start_time: _time.Seconds = None):
        self._cases: typing.List[TestCaseEntry] = []
        self.start_time = start_time

    def get_test_case(self) -> TestCase:
        try:
            return self._cases[-1].case
        except IndexError:
            return DummyTestCase()

    def get_parent_test_case(self) -> typing.Optional[TestCase]:
        try:
            return self._cases[-2].case
        except IndexError:
            return None

    def pop_test_case(self) -> TestCase:
        entry = self._cases.pop()
        elapsed_time = _time.time() - entry.start_time
        LOG.debug(f"Exit test case '{entry.case.id()}' after "
                  f"{elapsed_time} seconds")
        return entry.case

    def push_test_case(self, case: TestCase) -> TestCase:
        case = _exception.check_valid_type(case, TestCase)
        entry = TestCaseEntry(case=case,
                              start_time=_time.time())
        parent = self.get_test_case()
        self._cases.append(entry)
        LOG.debug(f"Enter test case '{case.id()}'")
        return parent


TEST_CASE_MANAGER = TestCaseManager()


def test_case_manager(manager: TestCaseManager = None) -> TestCaseManager:
    if manager is None:
        return TEST_CASE_MANAGER
    else:
        return tobiko.check_valid_type(manager, TestCaseManager)


def push_test_case(case: TestCase,
                   manager: TestCaseManager = None) -> TestCase:
    manager = test_case_manager(manager)
    return manager.push_test_case(case=case)


def pop_test_case(manager: TestCaseManager = None) -> TestCase:
    manager = test_case_manager(manager)
    return manager.pop_test_case()


def get_test_case(manager: TestCaseManager = None) -> TestCase:
    manager = test_case_manager(manager)
    return manager.get_test_case()


def get_parent_test_case(manager: TestCaseManager = None) \
        -> typing.Optional[TestCase]:
    manager = test_case_manager(manager)
    return manager.get_parent_test_case()


@contextlib.contextmanager
def enter_test_case(case: TestCase,
                    manager: TestCaseManager = None):
    manager = test_case_manager(manager)
    parent = manager.push_test_case(case)
    try:
        with parent.subTest(case.id()):
            yield
    finally:
        assert case is manager.pop_test_case()
        tobiko.remove_test_from_all_shared_resources(case.id())


def test_case(case: TestCase = None,
              manager: TestCaseManager = None) -> TestCase:
    if case is None:
        case = get_test_case(manager=manager)
    return _exception.check_valid_type(case, TestCase)


def get_sub_test_id(case: TestCase = None,
                    manager: TestCaseManager = None) -> str:
    # pylint: disable=protected-access
    case = test_case(case=case, manager=manager)
    if case._subtest is None:  # type: ignore
        return case.id()
    else:
        return case._subtest.id()  # type: ignore


def get_test_result(case: TestCase = None,
                    manager: TestCaseManager = None) \
        -> TestResult:
    case = test_case(case=case, manager=manager)
    outcome = getattr(case, '_outcome', None)
    result = getattr(outcome, 'result', None)
    if result is None:
        return TestResult()
    else:
        return result


def test_result(result: TestResult = None,
                case: TestCase = None,
                manager: TestCaseManager = None) \
        -> TestResult:
    if result is None:
        result = get_test_result(case=case, manager=manager)
    return tobiko.check_valid_type(result, TestResult)


RunTestType = typing.Union[TestCase, TestSuite]


def run_test(case: RunTestType,
             manager: TestCaseManager = None,
             result: TestResult = None,
             check=True) -> TestResult:
    if result is None:
        if check:
            parent = get_test_case(manager=manager)
            forward = get_test_result(case=parent)
            result = ForwardTestResult(forward=forward,
                                       parent=parent)
        else:
            result = TestResult()

    case.run(result=result)
    return result


ExcInfoType = typing.Union[
    typing.Tuple[typing.Type[BaseException],
                 BaseException,
                 types.TracebackType],
    typing.Tuple[None, None, None]]


class ForwardTestResult(TestResult):

    def __init__(self,
                 forward: TestResult,
                 parent: TestCase,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.forward = forward
        self.parent = parent

    def startTest(self, test: TestCase):
        super().startTest(test)
        if hasattr(self.forward, 'startTest'):
            self.forward.startTest(test)

    def stopTest(self, test: TestCase):
        super().stopTest(test)
        if hasattr(self.forward, 'stopTest'):
            self.forward.stopTest(test)

    def addError(self, test: TestCase, err: ExcInfoType):
        super().addError(test, err)
        if hasattr(self.forward, 'addError'):
            self.forward.addError(test, err)
            # self.forward.addError(self.parent, err)

    def addFailure(self, test: TestCase, err: ExcInfoType):
        super().addFailure(test, err)
        if hasattr(self.forward, 'addFailure'):
            self.forward.addFailure(test, err)
            # self.forward.addFailure(self.parent, err)

    def addSubTest(self,
                   test: TestCase,
                   subtest: TestCase,
                   err: typing.Optional[ExcInfoType]):
        super().addSubTest(test, subtest, err)
        if hasattr(self.forward, 'addSubTest'):
            self.forward.addSubTest(test, subtest, err)

    def addSuccess(self, test: TestCase):
        super().addSuccess(test)
        if hasattr(self.forward, 'addSuccess'):
            self.forward.addSuccess(test)

    def addSkip(self, test, reason: str):
        super().addSkip(test, reason)
        if hasattr(self.forward, 'addSkip'):
            self.forward.addSkip(test, reason)

    def addExpectedFailure(self, test: TestCase, err: ExcInfoType):
        super().addExpectedFailure(test, err)
        if hasattr(self.forward, 'addExpectedFailure'):
            self.forward.addExpectedFailure(test, err)

    def addUnexpectedSuccess(self, test: TestCase):
        super().addUnexpectedSuccess(test)
        if hasattr(self.forward, 'addUnexpectedSuccess'):
            self.forward.addUnexpectedSuccess(test)


def assert_in(needle, haystack,
              message: str = None):
    case = get_test_case()
    case.assertIn(needle, haystack, message)


def get_skipped_test_cases(skip_reason: str = None,
                           result: TestResult = None,
                           case: TestCase = None,
                           manager: TestCaseManager = None) \
        -> typing.List[TestCase]:
    result = test_result(result=result, case=case, manager=manager)
    return [case
            for case, reason in result.skipped
            if skip_reason is None or skip_reason in reason]


def assert_test_case_was_skipped(needle: TestCase,
                                 skip_reason: str = None,
                                 result: TestResult = None,
                                 case: TestCase = None,
                                 manager: TestCaseManager = None):
    skipped = get_skipped_test_cases(skip_reason=skip_reason,
                                     result=result,
                                     case=case,
                                     manager=manager)
    assert_in(needle, skipped)


FailureException = typing.cast(
    typing.Tuple[Exception, ...],
    (unittest.TestCase.failureException,
     testtools.TestCase.failureException,
     AssertionError))


def failure_exception_type(case: TestCase = None,
                           manager: TestCaseManager = None) \
        -> typing.Type[Exception]:
    case = test_case(case=case, manager=manager)
    assert issubclass(case.failureException, Exception)
    return case.failureException


def fail(msg: str,
         cause: typing.Type[Exception] = None) -> typing.NoReturn:
    """Fail immediately current test case execution, with the given message.

    Unconditionally raises a tobiko.FailureException as in below equivalent
    code:

        raise FailureException(msg.format(*args, **kwargs))

    :param msg: string message used to create FailureException
    :param cause: error that caused the failure
    :returns: It never returns
    :raises failure_type or FailureException exception type:
    """
    failure_type = failure_exception_type()
    raise failure_type(msg) from cause


def add_cleanup(function: typing.Callable, *args, **kwargs):
    get_test_case().addCleanup(function, *args, **kwargs)


def test_id(case: TestCase = None,
            manager: TestCaseManager = None) \
        -> str:
    return test_case(case=case, manager=manager).id()


def sub_test(msg: str = None, **kwargs):
    case = get_test_case()
    return case.subTest(msg, **kwargs)


def setup_tobiko_config(conf):
    # pylint: disable=unused-argument
    unittest.TestCase = BaseTestCase
    testtools.TestCase = TestToolsTestCase


def retry_test_case(*exceptions: Exception,
                    count: int = None,
                    timeout: _time.Seconds = None,
                    sleep_time: _time.Seconds = None,
                    interval: _time.Seconds = None) -> \
                    typing.Callable[[typing.Callable], typing.Callable]:
    """Re-run test case method in case it fails
    """
    if not exceptions:
        exceptions = FailureException
    return _retry.retry_on_exception(*exceptions,
                                     count=count,
                                     timeout=timeout,
                                     sleep_time=sleep_time,
                                     interval=interval,
                                     default_count=3,
                                     on_exception=on_test_case_retry_exception)


def on_test_case_retry_exception(attempt: _retry.RetryAttempt,
                                 case: testtools.TestCase,
                                 *_args, **_kwargs):
    if isinstance(case, testtools.TestCase):
        # pylint: disable=protected-access
        case._report_traceback(sys.exc_info(),
                               f"traceback[attempt={attempt.number}]")
    LOG.exception("Re-run test after failed attempt. "
                  f"(attempt={attempt.number}, test='{case.id()}')")



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import contextlib
import io
import os
import tempfile
import time
import typing  # noqa

from oslo_log import log
import requests

import tobiko
from tobiko.config import get_bool_env
from tobiko.openstack.glance import _client
from tobiko.openstack.glance import _io
from tobiko.openstack import keystone
from tobiko.shell import sh


LOG = log.getLogger(__name__)


class HasImageMixin(_client.HasGlanceClientMixin):

    @property
    def image_id(self):
        raise NotImplementedError

    @property
    def image(self):
        return self.get_image(image_id=self.image_id)


class GlanceImageStatus(object):

    #: The Image service reserved an image ID for the image in the catalog but
    # did not yet upload any image data.
    QUEUED = u'queued'

    #: The Image service is in the process of saving the raw data for the
    # image into the backing store.
    SAVING = u'saving'

    #: The image is active and ready for consumption in the Image service.
    ACTIVE = u'active'

    #: An image data upload error occurred.
    KILLED = u'killed'

    #: The Image service retains information about the image but the image is
    # no longer available for use.
    DELETED = u'deleted'

    #: Similar to the deleted status. An image in this state is not
    # recoverable.
    PENDING_DELETE = u'pending_delete'

    #: The image data is not available for use.
    DEACTIVATED = u'deactivated'

    #: Data has been staged as part of the interoperable image import process.
    # It is not yet available for use. (Since Image API 2.6)
    UPLOADING = u'uploading'

    #: The image data is being processed as part of the interoperable image
    # import process, but is not yet available for use. (Since Image API 2.6)
    IMPORTING = u'importing'


@keystone.skip_unless_has_keystone_credentials()
class GlanceImageFixture(_client.HasGlanceClientMixin, tobiko.SharedFixture):

    image_name: typing.Optional[str] = None
    username: typing.Optional[str] = None
    password: typing.Optional[str] = None
    disabled_algorithms: typing.Optional[typing.Dict[str, typing.Any]] = None
    image = None
    wait_interval = 5.

    def __init__(self,
                 image_name: typing.Optional[str] = None,
                 username: typing.Optional[str] = None, password:
                 typing.Optional[str] = None):
        super(GlanceImageFixture, self).__init__()

        if image_name:
            self.image_name = image_name
        elif not self.image_name:
            self.image_name = self.fixture_name
        tobiko.check_valid_type(self.image_name, str)

        if username:
            self.username = username

        if password:
            self.password = password

    def setup_fixture(self):
        self.setup_image()

    def cleanup_fixture(self):
        self.delete_image()

    def setup_image(self):
        return self.wait_for_image_active()

    def get_image(self, image_id=None, **params):
        if image_id or params:
            image = super(GlanceImageFixture, self).get_image(
                image_id=image_id, **params)
        else:
            image = self.find_image(name=self.image_name, default=None)
        self.image = image
        return image

    def delete_image(self, image_id=None):
        if not image_id:
            image_id = self.image_id
            self.image = None
        LOG.debug('Deleting Glance image %r (%r)...', self.image_name,
                  image_id)
        if _client.delete_image(image_id=image_id, client=self.glance_client):
            LOG.debug('Deleted Glance image %r (%r).', self.image_name,
                      image_id)

    @property
    def image_id(self):
        image = self.image or self.get_image()
        return image and image.id or None

    @property
    def image_status(self):
        image = self.image or self.get_image()
        return image and image.status or GlanceImageStatus.DELETED

    def wait_for_image_active(self, check=True):
        return self.wait_for_image_status(
            expected_status={GlanceImageStatus.ACTIVE}, check=check)

    def wait_for_image_deleted(self, check=True):
        return self.wait_for_image_status(
            expected_status={GlanceImageStatus.DELETED}, check=check)

    def wait_for_getting_active_image(self, check=True):
        return self.wait_for_image_status(
            expected_status=GETTING_ACTIVE_STATUS, check=check)

    def wait_for_image_status(self, expected_status, check=True):
        """Waits for the image to reach the given status."""
        image = self.image or self.get_image()
        while (image and image.status not in expected_status and
               is_image_status_changing(image)):

            LOG.debug("Waiting for %r (id=%r) stack status "
                      "(observed=%r, expected=%r)", self.image_name,
                      image.id, image.status, expected_status)
            time.sleep(self.wait_interval)
            image = self.get_image()

        if check:
            self.check_image_status(image, expected_status)
        return image

    def check_image_status(self, image, expected_status):
        image_status = image and image.status or GlanceImageStatus.DELETED
        if image_status not in expected_status:
            image_id = image and image.id or None
            raise InvalidGlanceImageStatus(image_name=self.image_name,
                                           image_id=image_id,
                                           actual_status=image_status,
                                           expected_status=expected_status)


class UploadGranceImageFixture(GlanceImageFixture):

    disk_format = "raw"
    container_format = "bare"
    create_image_retries = None
    tags: typing.List[str] = []

    def __init__(self, disk_format=None, container_format=None, tags=None,
                 **kwargs):
        super(UploadGranceImageFixture, self).__init__(**kwargs)

        if container_format:
            self.container_format = container_format
        tobiko.check_valid_type(self.container_format, str)

        if disk_format:
            self.disk_format = disk_format
        tobiko.check_valid_type(self.disk_format, str)

        self.tags = list(tags or self.tags)
        tobiko.check_valid_type(self.tags, list)

        self.prevent_image_create = get_bool_env('TOBIKO_PREVENT_CREATE')

    def setup_image(self):
        self.create_image()

    def create_image(self, retries=None):
        with self._cleanup_image_ids() as cleanup_image_ids:
            retries = retries or self.create_image_retries or 1
            while retries >= 0:
                image = self.wait_for_getting_active_image(
                    check=(not retries))
                if is_image_getting_active(image):
                    break
                retries -= 1

                if not self.prevent_image_create:
                    if image:
                        LOG.debug('Delete existing image: %r (id=%r)',
                                  self.image_name, image.id)
                        self.delete_image(image.id)
                        self.wait_for_image_deleted()

                    # Cleanup cached objects
                    self.image = image = None

                    LOG.debug('Creating Glance image %r '
                              '(re-tries left %d)...',
                              self.image_name, retries)
                    image_id = _client.create_image(
                        **self.create_image_parameters)['id']

                    cleanup_image_ids.add(image_id)
                    LOG.debug('Created image %r (id=%r)...',
                              self.image_name, image_id)

            if image:
                if image.id in cleanup_image_ids:
                    LOG.debug('Image created: %r (id=%r)',
                              self.image_name, image.id)
                    self.upload_image()
                else:
                    LOG.debug('Existing image found: %r (id=%r)',
                              self.image_name, image.id)

            image = self.wait_for_image_active()
            if image and image.id in cleanup_image_ids:
                # Having an active image we can remove image created from this
                # process from cleanup image list so that it is not going to
                # be deleted
                cleanup_image_ids.remove(image.id)

        return image

    @property
    def create_image_parameters(self):
        return dict(client=self.glance_client,
                    name=self.image_name,
                    disk_format=self.disk_format,
                    container_format=self.container_format,
                    tags=self.tags)

    @contextlib.contextmanager
    def _cleanup_image_ids(self):
        created_image_ids = set()
        try:
            yield created_image_ids
        finally:
            for image_id in created_image_ids:
                LOG.warning("Delete duplicate image %r (id=%r).",
                            self.image_name, image_id)
                try:
                    self.delete_image(image_id)
                except Exception:
                    LOG.exception('Error deleting image %r (%r)',
                                  self.image_name, image_id)

    def upload_image(self):
        self.check_image_status(self.image, {GlanceImageStatus.QUEUED})
        image_data, image_size = self.get_image_data()
        with image_data:
            _client.upload_image(image_id=self.image_id,
                                 image_data=image_data,
                                 image_size=image_size)
            LOG.debug("Image uploaded %r", self.image_name)

    def get_image_data(self):
        raise NotImplementedError


class FileGlanceImageFixture(UploadGranceImageFixture):

    image_file = None
    image_dir = None
    compression_type = None

    def __init__(self, image_file=None, image_dir=None, **kwargs):
        super(FileGlanceImageFixture, self).__init__(**kwargs)

        if image_file:
            self.image_file = image_file
        elif not self.image_file:
            self.image_file = self.fixture_name
        tobiko.check_valid_type(self.image_file, str)

        if image_dir:
            self.image_dir = image_dir
        elif not self.image_dir:
            from tobiko import config
            CONF = config.CONF
            self.image_dir = CONF.tobiko.glance.image_dir or "."
        tobiko.check_valid_type(self.image_dir, str)

    @property
    def real_image_dir(self):
        return os.path.realpath(os.path.expanduser(self.image_dir))

    @property
    def real_image_file(self):
        return os.path.join(self.real_image_dir, self.image_file)

    def get_image_data(self):
        return self.get_image_file(image_file=self.real_image_file)

    def get_image_file(self, image_file: str):
        image_size = os.path.getsize(image_file)
        LOG.debug('Uploading image %r data from file %r (%d bytes)',
                  self.image_name, image_file, image_size)
        image_data = _io.open_image_file(
            filename=image_file, mode='rb',
            compression_type=self.compression_type)
        return image_data, image_size


class URLGlanceImageFixture(FileGlanceImageFixture):

    image_url: str

    def __init__(self,
                 image_url: str = None,
                 **kwargs):
        super().__init__(**kwargs)
        if image_url is None:
            image_url = self.image_url
        else:
            self.image_url = image_url
        tobiko.check_valid_type(image_url, str)

    def get_image_file(self, image_file: str):
        http_request = requests.get(self.image_url, stream=True)
        expected_size = int(http_request.headers.get('content-length', 0))
        chunks = http_request.iter_content(chunk_size=io.DEFAULT_BUFFER_SIZE)
        try:
            http_request.raise_for_status()
        except requests.exceptions.HTTPError:
            error_content = b''.join(chunks).decode()
            LOG.exception("Error getting image file from URL: "
                          f"{self.image_url}:\n"
                          f"{error_content}")
            raise

        download_image = True
        if expected_size:
            try:
                actual_size = os.path.getsize(image_file)
            except Exception as ex:
                LOG.debug("Unable to get image %r data from file %r: %s",
                          self.image_name, image_file, ex)
            else:
                if actual_size == expected_size:
                    LOG.debug("Cached image %r file %r found (%d bytes)",
                              self.image_name, image_file, actual_size)
                    download_image = False

        if download_image:
            LOG.debug("Downloading image %r from URL %r to file %r "
                      "(%d bytes)", self.image_name, self.image_url,
                      image_file, expected_size)
            self._download_image_file(image_file=image_file,
                                      chunks=chunks,
                                      expected_size=expected_size)
        image_file = self.customize_image_file(base_file=image_file)
        return super(URLGlanceImageFixture, self).get_image_file(
            image_file=image_file)

    def _download_image_file(self, image_file, chunks, expected_size):
        image_dir = os.path.dirname(image_file)
        LOG.debug('Ensure image directory exists: %r', image_dir)
        tobiko.makedirs(image_dir)

        fd, temp_file = tempfile.mkstemp(dir=image_dir)
        with io.open(fd, 'wb', io.DEFAULT_BUFFER_SIZE) as image_data:
            for chunk in chunks:
                image_data.write(chunk)

        actual_size = os.path.getsize(temp_file)
        LOG.debug('Downloaded image %r from URL %r to file %r (%d bytes)',
                  self.image_name, self.image_url, image_file,
                  actual_size)

        if expected_size and actual_size != expected_size:
            message = "Download file size mismatch: {!s} != {!r}".format(
                expected_size, actual_size)
            raise RuntimeError(message)
        os.rename(temp_file, image_file)

    def customize_image_file(self, base_file: str) -> str:
        return base_file


class CustomizedGlanceImageFixture(URLGlanceImageFixture):

    @property
    def firstboot_commands(self) -> typing.List[str]:
        return []

    @property
    def install_packages(self) -> typing.List[str]:
        return []

    @property
    def run_commands(self) -> typing.List[str]:
        return []

    @property
    def write_files(self) -> typing.Dict[str, str]:
        return {}

    username: str = ''
    password: str = ''

    def _get_customized_suffix(self) -> str:
        return 'customized'

    def customize_image_file(self, base_file: str) -> str:
        customized_file = f'{base_file}-{self._get_customized_suffix()}'
        if os.path.isfile(customized_file):
            if (os.stat(base_file).st_mtime_ns <
                    os.stat(customized_file).st_mtime_ns):
                LOG.debug(f"Image file is up to date '{customized_file}'")
                return customized_file
            else:
                LOG.debug(f"Remove obsolete image file '{customized_file}'")
                os.remove(customized_file)
        work_file = sh.execute('mktemp').stdout.strip()
        try:
            LOG.debug(f"Copy base image file: '{base_file}' to '{work_file}'")
            sh.put_file(base_file, work_file)

            options = self.get_virt_customize_options()
            if options:
                command = sh.shell_command(['virt-customize', '-a', work_file])
                sh.execute(command + options)

            sh.get_file(work_file, customized_file)
            return customized_file
        finally:
            sh.execute(['rm', '-f', work_file])

    def get_virt_customize_options(self) -> sh.ShellCommand:
        options = sh.ShellCommand()

        firstboot_commands = self.firstboot_commands
        if firstboot_commands:
            for cmd in firstboot_commands:
                options += ['--firstboot-command', cmd]

        install_packages = self.install_packages
        if install_packages:
            options += ['--install', ','.join(install_packages)]

        username = self.username
        password = self.password
        if username and password:
            options += f'--password "{username}:password:{password}"'

        run_commands = self.run_commands
        if run_commands:
            for cmd in run_commands:
                options += ['--run-command', cmd]

        write_files = self.write_files
        if write_files:
            for filename, content in write_files.items():
                options += ['--write', f'{filename}:{content}']

        return options


class InvalidGlanceImageStatus(tobiko.TobikoException):
    message = ("Invalid image {image_name!r} (id {image_id!r}) status: "
               "{actual_status!r} not in {expected_status!r}")


class GlanceImageCreationFailed(tobiko.TobikoException):
    message = ("Failed creating image {image_name!r}: status "
               "({observed!r}) not in ({expected!r})")


CHANGING_STATUS = {GlanceImageStatus.QUEUED,
                   GlanceImageStatus.IMPORTING,
                   GlanceImageStatus.PENDING_DELETE,
                   GlanceImageStatus.SAVING,
                   GlanceImageStatus.UPLOADING}


def is_image_status_changing(image):
    return image and image.status in CHANGING_STATUS


GETTING_ACTIVE_STATUS = {GlanceImageStatus.QUEUED,
                         GlanceImageStatus.SAVING,
                         GlanceImageStatus.ACTIVE}


def is_image_getting_active(image):
    return image and image.status in GETTING_ACTIVE_STATUS



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg


GROUP_NAME = 'tripleo'
OPTIONS = [
    # Undercloud options
    cfg.StrOpt('undercloud_ssh_hostname',
               default='undercloud-0',
               help="hostname or IP address to be used to connect to "
                    "undercloud host"),
    cfg.IntOpt('undercloud_ssh_port',
               default=None,
               help="TCP port of SSH server on undercloud host"),
    cfg.StrOpt('undercloud_ssh_username',
               default='stack',
               help="Username with access to stackrc and overcloudrc files"),
    cfg.StrOpt('undercloud_ssh_key_filename',
               default=None,
               help="SSH key filename used to login to Undercloud node"),
    cfg.ListOpt('undercloud_rcfile',
                default=['~/stackrc'],
                help="Undercloud RC filename"),
    cfg.StrOpt('undercloud_cloud_name',
               default='undercloud',
               help='undercloud cloud name to be used for loading credentials '
                    'from the undercloud clouds files'),


    # Overcloud options
    cfg.IntOpt('overcloud_ssh_port',
               default=None,
               help="TCP port of SSH server on overcloud hosts"),
    cfg.StrOpt('overcloud_ssh_username',
               default='heat-admin',
               help="Default username used to connect to overcloud nodes"),
    cfg.StrOpt('overcloud_ssh_key_filename',
               default='~/.ssh/id_overcloud',
               help="SSH key filename used to login to Overcloud nodes"),
    cfg.ListOpt('overcloud_rcfile',
                default=['~/overcloudrc', '~/qe-Cloud-0rc'],
                help="Overcloud RC filenames"),
    cfg.StrOpt('overcloud_cloud_name',
               default='overcloud',
               help='overcloud cloud name to be used for loading credentials '
                    'from the overcloud clouds files'),
    cfg.IntOpt('overcloud_ip_version',
               help=("Default IP address version to be used to connect to "
                     "overcloud nodes ")),
    cfg.StrOpt('overcloud_network_name',
               help="Name of network used to connect to overcloud nodes"),
    cfg.DictOpt('overcloud_groups_dict',
                help='Dictionary with the node groups corresponding to '
                     'different hostname prefixes',
                default={'ctrl': 'controller', 'cmp': 'compute'}),

    # Other options
    cfg.StrOpt('inventory_file',
               default='.ansible/inventory/tripleo.yaml',
               help="path to where to export tripleo inventory file"),

    cfg.BoolOpt('has_external_load_balancer',
                default=False,
                help="OSP env was done with an external load balancer"),

    cfg.BoolOpt('ceph_rgw',
                default=False,
                help="whether Ceph RGW is deployed"),
]


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup(GROUP_NAME), opts=OPTIONS)


def list_options():
    return [(GROUP_NAME, itertools.chain(OPTIONS))]


def setup_tobiko_config(conf):
    # pylint: disable=unused-argument
    from tobiko.tripleo import _ansible
    from tobiko.tripleo import overcloud
    from tobiko.tripleo import topology

    _ansible.setup_undercloud_ansible_playbook()
    overcloud.setup_overcloud_keystone_credentials()
    topology.setup_tripleo_topology()



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import socket

import testtools

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.openstack import keystone
from tobiko.openstack import stacks


class GetHostnameTest(testtools.TestCase):

    def test_hostname(self,
                      expect_hostname: str = None,
                      **execute_params):
        hostname = sh.get_hostname(**execute_params)
        self.assertIsInstance(hostname, str)
        if expect_hostname:
            self.assertEqual(expect_hostname, hostname)
        else:
            self.assertNotEqual('', hostname)

    def test_local_hostname(self):
        self.test_hostname(expect_hostname=socket.gethostname(),
                           ssh_client=False)

    def test_ssh_hostname(self,
                          ssh_client: ssh.SSHClientFixture = None):
        fixture = ssh.ssh_client_fixture(ssh_client)
        if fixture is None:
            expect_hostname = socket.gethostname()
        else:
            stdin, stdput, stderr = fixture.connect().exec_command('hostname')
            stdin.close()
            self.assertEqual(b'', stderr.read())
            expect_hostname = stdput.read().decode().strip()
        self.test_hostname(ssh_client=ssh_client,
                           expect_hostname=expect_hostname)

    def test_ssh_proxy_hostname(self):
        ssh_client = ssh.ssh_proxy_client()
        if ssh_client is None:
            tobiko.skip_test('SSH proxy server is not configured')
        self.test_ssh_hostname(ssh_client=ssh_client)

    cirros_server = tobiko.required_fixture(
        stacks.CirrosServerStackFixture)

    @keystone.skip_unless_has_keystone_credentials()
    def test_cirros_hostname(self):
        self.test_ssh_hostname(ssh_client=self.cirros_server.ssh_client)

    ubuntu_server = tobiko.required_fixture(
        stacks.UbuntuServerStackFixture)

    @keystone.skip_unless_has_keystone_credentials()
    def test_ubuntu_hostname(self):
        self.test_ssh_hostname(ssh_client=self.ubuntu_server.ssh_client)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.actors import _actor
from tobiko.actors import _manager
from tobiko.actors import _proxy

actor_method = _actor.actor_method
Actor = _actor.Actor
ActorRef = _actor.ActorRef

cleanup_actor = _manager.cleanup_actor
setup_actor = _manager.setup_actor
start_actor = _manager.start_actor
stop_actor = _manager.stop_actor

call_proxy = _proxy.create_call_proxy
call_proxy_class = _proxy.create_call_proxy_class
CallProxy = _proxy.CallProxy
CallProxyBase = _proxy.CallProxyBase



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from oslo_log import log

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack.nova import _client


LOG = log.getLogger(__name__)


def get_nova_quota_set(project: keystone.ProjectType = None,
                       user: keystone.UserType = None,
                       client: _client.NovaClientType = None,
                       **params):
    client = _client.nova_client(client)
    project = keystone.get_project_id(project=project,
                                      session=client.client.session)
    user = user and keystone.get_user_id(user=user) or None
    return client.quotas.get(project, user_id=user, **params)


def set_nova_quota_set(project: keystone.ProjectType = None,
                       user: keystone.UserType = None,
                       client: _client.NovaClientType = None,
                       **params):
    client = _client.nova_client(client)
    project = keystone.get_project_id(project=project,
                                      session=client.client.session)
    user = user and keystone.get_user_id(user=user) or None
    return client.quotas.update(project, user_id=user, **params)


def ensure_nova_quota_limits(project: keystone.ProjectType = None,
                             user: keystone.UserType = None,
                             client: _client.NovaClientType = None,
                             retry_timeout: tobiko.Seconds = None,
                             retry_interval: tobiko.Seconds = None,
                             **required_quotas: int):
    if not required_quotas:
        return

    client = _client.nova_client(client)
    project = keystone.get_project_id(project=project,
                                      session=client.client.session)
    user = user and keystone.get_user_id(user=user) or None
    if user:
        # Must increase project limits before user ones
        ensure_nova_quota_limits(project=project, client=client,
                                 **required_quotas)

    for attempt in tobiko.retry(timeout=retry_timeout,
                                interval=retry_interval,
                                default_timeout=60.,
                                default_interval=3.):
        actual_limits, expected_limits = get_nova_quota_limits_increase(
            project=project, user=user, client=client,
            extra_increase=10//attempt.number, **required_quotas)
        if expected_limits:
            if attempt.is_last:
                raise EnsureNovaQuotaLimitsError(
                    project=project,
                    actual_limits=actual_limits,
                    expected_limits=expected_limits)
            LOG.info(f"Increase Nova quota limit(s) (project={project}, "
                     f"user={user}): {actual_limits} -> {expected_limits}...")
            try:
                set_nova_quota_set(project=project, user=user, client=client,
                                   **expected_limits)
            except Exception:
                if attempt.is_last:
                    raise
                LOG.exception("Error increasing Nova quota set limits: "
                              f"{expected_limits}")
        else:
            LOG.debug(f"Required Nova quota limits are OK: {required_quotas}")
            break
    else:
        raise RuntimeError("Broken retry loop")


class EnsureNovaQuotaLimitsError(tobiko.TobikoException):
    message = ("Neutron quota limits lower than "
               "expected (project={project}): "
               "{actual_limits} != {expected_limits}")


def get_nova_quota_limits_increase(
        project: keystone.ProjectType = None,
        user: keystone.UserType = None,
        client: _client.NovaClientType = None,
        extra_increase=0,
        **required_quotas: int) \
        -> typing.Tuple[typing.Dict[str, int],
                        typing.Dict[str, int]]:
    quota_set = get_nova_quota_set(project=project, user=user,
                                   client=client, detail=True)
    LOG.debug("Got Nova quota set:\n"
              f"{quota_set}")
    actual_limits: typing.Dict[str, int] = {}
    expected_limits: typing.Dict[str, int] = {}
    for name, needed in required_quotas.items():
        quota = getattr(quota_set, name)
        limit: int = int(quota['limit'])
        if limit >= 0:
            in_use = max(0, int(quota['in_use']))
            reserved = max(0, int(quota['reserved']))
            required_limit = in_use + reserved + needed
            if required_limit >= limit:
                actual_limits[name] = limit
                expected_limits[name] = required_limit + extra_increase
    return actual_limits, expected_limits



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import math
import typing  # noqa

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh


class FilesNotFound(tobiko.TobikoException):
    message = ("Files not found (path={path}, name={name}, login={login}, "
               "exit_status={exit_status}):\n{stderr}")


NameType = typing.Union[None, str, typing.List[str]]
PathType = typing.Union[str, typing.Iterable[str]]


def find_files(path: sh.ShellCommandType,
               name: NameType = None,
               command: sh.ShellCommandType = 'find',
               max_depth: int = None,
               modified_since: tobiko.Seconds = None,
               ssh_client: ssh.SSHClientType = None,
               type: str = None,
               **execute_params) -> typing.List[str]:
    # pylint: disable=redefined-builtin
    if not path:
        raise ValueError("Path can't be empty")

    command_line = sh.shell_command(command) + path

    if max_depth is not None:
        command_line += f"-maxdepth '{max_depth}'"

    if type is not None:
        command_line += f"-type '{type}'"

    if name is not None:
        if isinstance(name, str):
            names = [name]
        else:
            names = list(name)
        if names:
            command_line += ' -o '.join(f"-name '{name}'" for name in names)

    if modified_since is not None:
        # round seconds to the next minute
        minutes = math.ceil(modified_since / 60.)
        command_line += f"-mmin {minutes}"

    result = sh.execute(command_line,
                        ssh_client=ssh_client,
                        expect_exit_status=None,
                        **execute_params)
    file_names: typing.List[str] = [line.strip()
                                    for line in result.stdout.splitlines()
                                    if line.strip()]
    if not file_names:
        ssh_client = ssh.ssh_client_fixture(ssh_client)
        login: typing.Optional[str] = None
        if ssh_client is not None:
            login = ssh_client.login
        raise FilesNotFound(path=path,
                            name=name,
                            login=login,
                            exit_status=result.exit_status,
                            stderr=result.stderr.strip())
    return file_names



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import os
import shlex
import typing

from oslo_log import log

import tobiko
from tobiko.shell import grep
from tobiko.shell import find
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


class LogFileDigger(tobiko.SharedFixture):

    found: typing.MutableMapping[str, None]

    def __init__(self, filename: str,
                 pattern: typing.Optional[str] = None,
                 **execute_params):
        super(LogFileDigger, self).__init__()
        self.filename = filename
        self.pattern = pattern
        self.execute_params = execute_params
        self.found = collections.OrderedDict()

    def setup_fixture(self):
        if self.pattern is not None:
            self.find_lines()

    def cleanup_fixture(self):
        self.found.clear()

    @property
    def found_lines(self) -> typing.List[str]:
        return list(self.found)

    def find_lines(self,
                   pattern: str = None,
                   new_lines=False) \
            -> typing.List[str]:
        if pattern is None:
            pattern = self.pattern
            if pattern is None:
                raise ValueError(f"Invalid pattern: {pattern}")
        elif self.pattern is not None:
            raise NotImplementedError(
                "Combining patterns is not supported")
        try:
            lines = self.grep_lines(pattern,
                                    new_lines=new_lines)
        except grep.NoMatchingLinesFound:
            lines = []
        else:
            if lines:
                lines = [line
                         for line in lines
                         if line not in self.found]
                self.found.update((line, None)
                                  for line in lines)

        if new_lines:
            if lines:
                lines_text = '\n\t'.join(lines)
                LOG.debug('Found new lines:\n'
                          f"\t{lines_text}")
            return lines
        else:
            return list(self.found)

    def find_new_lines(self,
                       pattern: str = None) \
            -> typing.List[str]:
        return self.find_lines(pattern=pattern,
                               new_lines=True)

    def grep_lines(self,
                   pattern: str,
                   new_lines: bool = False) \
            -> typing.List[str]:
        # pylint: disable=unused-argument
        log_files = self.list_log_files()
        return grep.grep_files(pattern=pattern,
                               files=log_files,
                               **self.execute_params)

    def list_log_files(self):
        file_path, file_name = os.path.split(self.filename)
        return find.find_files(path=file_path,
                               name=file_name,
                               **self.execute_params)


class JournalLogDigger(LogFileDigger):

    def grep_lines(self,
                   pattern: str,
                   new_lines: bool = False) \
            -> typing.List[str]:
        try:
            result = sh.execute(["journalctl", '--no-pager',
                                 "--unit", shlex.quote(self.filename),
                                 # "--since", "30 minutes ago",
                                 '--output', 'short-iso',
                                 '--grep', shlex.quote(pattern)],
                                **self.execute_params)
        except sh.ShellCommandFailed as ex:
            if ex.stdout.endswith('-- No entries --\n'):
                ssh_client = self.execute_params.get('ssh_client')
                raise grep.NoMatchingLinesFound(
                    pattern=pattern,
                    files=[self.filename],
                    login=ssh_client and ssh_client.login or None)
            else:
                LOG.exception(f"Error executing journalctl: {ex.stderr}")
                return []
        else:
            lines = [line
                     for line in result.stdout.splitlines()
                     if not line.startswith('-- ')]
            return lines


class MultihostLogFileDigger(tobiko.SharedFixture):

    diggers: typing.Optional[typing.Dict[str, LogFileDigger]] = None

    def __init__(
            self,
            filename: str,
            ssh_clients: typing.Iterable[ssh.SSHClientType] = None,
            file_digger_class: typing.Type[LogFileDigger] = LogFileDigger,
            pattern: str = None,
            **execute_params):
        super(MultihostLogFileDigger, self).__init__()
        self.file_digger_class = file_digger_class
        self.filename = filename
        self.execute_params = execute_params
        self.pattern = pattern
        self.ssh_clients: typing.List[ssh.SSHClientType] = []
        if ssh_clients is not None:
            self.ssh_clients.extend(ssh_clients)

    def setup_fixture(self):
        for ssh_client in self.ssh_clients:
            self.add_host(ssh_client=ssh_client)
        if self.diggers is not None:
            for digger in self.diggers.values():
                self.useFixture(digger)

    def cleanup_fixture(self):
        self.diggers = None

    def add_host(self,
                 ssh_client: ssh.SSHClientType,
                 hostname: str = None):
        if self.diggers is None:
            self.diggers = collections.OrderedDict()
        if hostname is None:
            hostname = sh.get_hostname(ssh_client=ssh_client)
        digger = self.diggers.get(hostname)
        if digger is None:
            self.diggers[hostname] = digger = self.file_digger_class(
                filename=self.filename,
                ssh_client=ssh_client,
                pattern=self.pattern,
                **self.execute_params)
        return digger

    @property
    def found_lines(self) \
            -> typing.List[typing.Tuple[str, str]]:
        # ensure diggers are ready before looking for lines
        tobiko.setup_fixture(self)
        lines: typing.List[typing.Tuple[str, str]] = []
        if self.diggers is not None:
            for hostname, digger in self.diggers.items():
                for line in digger.found_lines:
                    lines.append((hostname, line))
        return lines

    def find_lines(self,
                   pattern: str = None,
                   new_lines: bool = False) \
            -> typing.List[typing.Tuple[str, str]]:
        # ensure diggers are ready before looking for lines
        tobiko.setup_fixture(self)
        lines: typing.List[typing.Tuple[str, str]] = []
        if self.diggers is not None:
            for hostname, digger in self.diggers.items():
                for line in digger.find_lines(pattern=pattern,
                                              new_lines=new_lines):
                    lines.append((hostname, line))
        return lines

    def find_new_lines(self,
                       pattern: str = None,
                       retry_count: int = None,
                       retry_timeout: tobiko.Seconds = 60.,
                       retry_interval: tobiko.Seconds = None) \
            -> typing.List[typing.Tuple[str, str]]:
        for _ in tobiko.retry(count=retry_count,
                              timeout=retry_timeout,
                              interval=retry_interval,
                              default_interval=1.,
                              default_timeout=60.):
            new_lines = self.find_lines(pattern=pattern,
                                        new_lines=True)
            if new_lines:
                break
        else:
            raise RuntimeError("Internal bug: retry loop break itself.")

        return new_lines



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import json
import typing

import tobiko
from tobiko.openstack.topology import _topology
from tobiko.shell import ip


def get_hosts_namespaces(hostnames: typing.Iterable[str] = None,
                         **params) \
        -> typing.Dict[str, typing.List[str]]:
    if isinstance(hostnames, str):
        hostnames = [hostnames]
    namespaces = collections.defaultdict(list)
    nodes = _topology.list_openstack_nodes(hostnames=hostnames,
                                           **params)
    for node in nodes:
        for namespace in ip.list_network_namespaces(
                ssh_client=node.ssh_client):
            namespaces[namespace].append(node.hostname)
    return namespaces


def wait_for_namespace_in_hosts(*namespaces: str,
                                hostnames: typing.Iterable[str] = None,
                                timeout: tobiko.Seconds = None,
                                count: int = None,
                                interval: tobiko.Seconds = None,
                                present=True,
                                **params):
    for attempt in tobiko.retry(timeout=timeout,
                                count=count,
                                interval=interval,
                                default_timeout=60.,
                                default_interval=5.):
        try:
            if present:
                assert_namespace_in_hosts(*namespaces,
                                          hostnames=hostnames,
                                          **params)
            else:
                assert_namespace_not_in_hosts(*namespaces,
                                              hostnames=hostnames,
                                              **params)
        except tobiko.FailureException:  # type: ignore
            if attempt.is_last:
                raise
        else:
            break


def assert_namespace_in_hosts(*namespaces: str,
                              hostnames: typing.Iterable[str] = None,
                              **params):
    actual_namespaces = get_hosts_namespaces(hostnames=hostnames,
                                             **params)
    missing = set(namespaces) - set(actual_namespaces)
    if missing:
        actual_hostnames = sorted(set(
            _hostname
            for _hostnames in actual_namespaces.values()
            for _hostname in _hostnames))
        tobiko.fail(f"Network namespace(s) {sorted(missing)} missing in "
                    f"host(s) {actual_hostnames!r}")


def assert_namespace_not_in_hosts(*namespaces: str,
                                  hostnames: typing.Iterable[str] = None,
                                  **params):
    unexpected_namespaces = collections.defaultdict(list)
    actual_namespaces = get_hosts_namespaces(hostnames=hostnames, **params)
    for namespace, hostnames in actual_namespaces.items():
        if namespace in sorted(set(namespaces)):
            for hostname in hostnames:
                unexpected_namespaces[hostname].append(namespace)
    if unexpected_namespaces:
        dump = json.dumps(unexpected_namespaces, indent=4, sort_keys=True)
        tobiko.fail(f"Unexpected network namespace(s) found in "
                    f"host(s):\n{dump}")



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from collections import abc
import functools
import socket
import typing

import netaddr
from oslo_log import log

import tobiko


LOG = log.getLogger(__name__)


def list_addresses(obj,
                   ip_version: typing.Optional[int] = None,
                   port: typing.Union[int, str, None] = None,
                   ssh_config: bool = False) -> \
        tobiko.Selection[netaddr.IPAddress]:
    if isinstance(obj, tobiko.Selection):
        addresses = obj
    elif isinstance(obj, netaddr.IPAddress):
        addresses = tobiko.select([obj])
    elif isinstance(obj, str):
        addresses = tobiko.select(
            list_host_addresses(obj,
                                ip_version=ip_version,
                                port=port,
                                ssh_config=ssh_config))
    elif isinstance(obj, abc.Sequence):
        addresses = tobiko.Selection()
        for item in iter(obj):
            addresses.extend(list_addresses(item))
    else:
        raise TypeError(f'{obj} object is not of a valid address(es) type')

    if addresses and ip_version is not None:
        addresses = addresses.with_attributes(version=ip_version)
    return addresses


@functools.lru_cache()
def list_host_addresses(host: str,
                        ip_version: typing.Optional[int] = None,
                        port: typing.Union[int, str, None] = None,
                        ssh_config: bool = False) -> \
        tobiko.Selection[netaddr.IPAddress]:

    if not port:
        if ssh_config:
            port = 22  # use the default port for SSH protocol
        else:
            port = 0

    addresses: tobiko.Selection[netaddr.IPAddress] = tobiko.Selection()
    hosts = [host]
    resolved = set()
    while hosts:
        host = hosts.pop()
        if host in resolved:
            LOG.debug(f"Cyclic address resolution detected for host {host}")
            continue  # already resolved

        resolved.add(host)  # avoid resolving it again
        address = parse_ip_address(host)
        if address:
            addresses.append(address)
            continue

        # use socket host address resolution to get IP addresses
        addresses.extend(resolv_host_addresses(host=host,
                                               port=port,
                                               ip_version=ip_version))

    if [host] != [str(address) for address in addresses]:
        LOG.debug(f"Host '{host}' addresses resolved as: {addresses}")
    return addresses


def parse_ip_address(host: str) -> typing.Optional[netaddr.IPAddress]:
    try:
        return netaddr.IPAddress(host)
    except (netaddr.AddrFormatError, ValueError):
        return None


ADDRESS_FAMILIES = {
    4: socket.AF_INET,
    6: socket.AF_INET6,
    None: socket.AF_UNSPEC
}


# pylint: disable=no-member
AddressFamily = socket.AddressFamily
# pylint: enable=no-member


def get_address_family(ip_version: typing.Optional[int] = None) -> \
        AddressFamily:
    try:
        return ADDRESS_FAMILIES[ip_version]
    except KeyError:
        pass
    raise ValueError(f"{ip_version!r} is an invalid value for getting address "
                     "family")


IP_VERSIONS = {
    socket.AF_INET: 4,
    socket.AF_INET6: 6,
}


def get_ip_version(family: AddressFamily) -> int:
    try:
        return IP_VERSIONS[family]
    except KeyError:
        pass
    raise ValueError(f"{family!r} is an invalid value for getting IP version")


def resolv_host_addresses(host: str,
                          port: typing.Union[int, str] = 0,
                          ip_version: typing.Optional[int] = None) -> \
        typing.List[netaddr.IPAddress]:

    family = get_address_family(ip_version)
    proto = socket.AI_CANONNAME | socket.IPPROTO_TCP
    LOG.debug(f"Resolve IP addresses for host '{host}' "
              f"(port={port}, family={family}, proto={proto})'...")
    try:
        addrinfo = socket.getaddrinfo(host, port, family=family, proto=proto)
    except socket.gaierror as ex:
        LOG.debug(f"Can't resolve IP addresses for host '{host}': {ex}")
        return []

    addresses = []
    for _family, _, _, canonical_name, sockaddr in addrinfo:
        if family != socket.AF_UNSPEC and family != _family:
            LOG.error(f"Resolved address family '{_family}' 'of address "
                      f"'{sockaddr}' is not {family} "
                      f"(canonical_name={canonical_name}")
            continue

        address = parse_ip_address(sockaddr[0])
        if address is None:
            LOG.error(f"Resolved address '{sockaddr[0]}' is not a valid IP "
                      f"address (canonical_name={canonical_name})")
            continue

        if ip_version and ip_version != address.version:
            LOG.error(f"Resolved IP address version '{address.version}' of "
                      f"'{address}' is not {ip_version} "
                      f"(canonical_name={canonical_name})")
            continue

        addresses.append(address)
        LOG.debug(f"IP address for host '{host}' has been resolved as "
                  f"'{address}' (canonical_name={canonical_name})")

    if not addresses:
        LOG.debug(f"Host name '{host}' resolved to any IP address.")

    return addresses



# Copyright (c) 2023 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json

from oslo_log import log
import testtools

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import stacks
from tobiko.openstack import topology
from tobiko.shell import sh
from tobiko.tripleo import containers
from tobiko.tripleo import overcloud

LOG = log.getLogger(__name__)


class BaseSecurityGroupTest(testtools.TestCase):

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(
        stacks.CirrosServerWithDefaultSecurityGroupStackFixture)

    _ovn_nb_db = None
    _host_ssh_client = None
    _container_runtime_name = None
    _container_name = None

    def setUp(self):
        super(BaseSecurityGroupTest, self).setUp()
        self.ovn_controller_agents = neutron.list_networking_agents(
            binary=neutron.OVN_CONTROLLER)
        if len(self.ovn_controller_agents) < 1:
            self.skip(f"No running {neutron.OVN_CONTROLLER} agents found. "
                      f"Stateless Security Group tests requires ML2/OVN "
                      f"deployment.")

    @property
    def ovn_nb_db(self):
        if not self._ovn_nb_db:
            command_result = sh.execute(
                "ovs-vsctl get open . external_ids:ovn-remote | "
                "sed -e 's/\"//g' | sed 's/6642/6641/g'",
                ssh_client=self.host_ssh_client,
                sudo=True)
            self._ovn_nb_db = command_result.stdout
        return self._ovn_nb_db

    @property
    def host_ssh_client(self):
        if not self._host_ssh_client:
            self._host_ssh_client = topology.get_openstack_node(
                hostname=self.ovn_controller_agents[0]['host']).ssh_client
        return self._host_ssh_client

    @property
    def container_runtime_name(self):
        if not self._container_runtime_name:
            if overcloud.has_overcloud():
                self._container_runtime_name = (
                    containers.get_container_runtime_name())
            else:
                self._container_runtime_name = 'docker'
        return self._container_runtime_name

    @property
    def container_name(self):
        if self._container_name is None:
            os_topology = topology.get_openstack_topology()
            if os_topology.has_containers:
                self._container_name = topology.get_agent_container_name(
                    neutron.OVN_CONTROLLER
                )
            else:
                self._container_name = ""
        return self._container_name

    def _check_sg_rule_in_ovn_nb_db(self, rule_id: str, expected_action: str):
        os_topology = topology.get_openstack_topology()
        command = ""
        if os_topology.has_containers:
            command += (
                f"{self.container_runtime_name} exec {self.container_name} ")
        command += (
            f"ovn-nbctl --format json --no-leader-only --db={self.ovn_nb_db} "
            f"find ACL external_ids:\"neutron\\:security_group_rule_id\"="
            f"\"{rule_id}\""
        )
        command_result = sh.execute(
            command, ssh_client=self.host_ssh_client, sudo=True)
        acl_rule = json.loads(command_result.stdout)
        self._assert_acl_action(acl_rule, expected_action)

    def _assert_acl_action(self, acl: dict, expected_action: str):
        action_column = acl['headings'].index('action')
        self.assertEqual(
            expected_action,
            acl['data'][0][action_column])

    def _check_sg_rules_in_ovn_nb_db(self, sg: dict, expected_action: str):
        sg_rules_ids = [
            rule['id'] for rule in sg['security_group_rules']]
        for sg_rule_id in sg_rules_ids:
            self._check_sg_rule_in_ovn_nb_db(sg_rule_id, expected_action)


@neutron.skip_if_missing_networking_extensions('stateful-security-group')
class StatelessSecurityGroupTest(BaseSecurityGroupTest):

    def test_default_security_group_is_stateful(self):
        """Test that default security group is always stateful.

        This test checks if default SG created for the project is stateful
        and if OVN's ACLs corresponding to the SG's rules have correct
        action which is "allow-related".

        Steps:
        1. Get default SG for the project (it's always created automatically),
        2. Check if ACLs corresponding to the rules from that SG have
           "action-related" action,
        3. Add new SG rule in the default SG,
        4. Check action of the ACL corresponding to the newly created SG rule.
        """

        default_sg = neutron.get_default_security_group(
            project_id=self.stack.project)
        self.assertTrue(default_sg['stateful'])
        self._check_sg_rules_in_ovn_nb_db(default_sg,
                                          neutron.STATEFUL_OVN_ACTION)
        new_rule = neutron.create_security_group_rule(
            default_sg['id'],
            port_range_min=1111,
            port_range_max=1111,
            ethertype="IPv4",
            protocol="tcp",
            description="test_default_security_group_is_stateful rule",
            direction="ingress"
        )
        self._check_sg_rule_in_ovn_nb_db(new_rule['id'],
                                         neutron.STATEFUL_OVN_ACTION)

    def test_new_security_group_is_stateful(self):
        """Test that newly created security group is stateful by default.

        This test checks if newly created SG is stateful by default
        and if OVN's ACLs corresponding to the SG's rules have correct
        action which is "allow-related".

        Steps:
        1. Create SG for the project,
        2. Check if ACLs corresponding to the rules from that SG have
           "action-related" action,
        3. Add new SG rule in the SG,
        4. Check action of the ACL corresponding to the newly created SG rule.
        """
        sg = neutron.create_security_group(
            name="test_new_security_group_is_statefull_SG",
        )
        self.assertTrue(sg['stateful'])
        self._check_sg_rules_in_ovn_nb_db(sg, neutron.STATEFUL_OVN_ACTION)
        new_rule = neutron.create_security_group_rule(
            sg['id'],
            port_range_min=1111,
            port_range_max=1111,
            ethertype="IPv4",
            protocol="tcp",
            description="test_new_security_group_is_statefull_SG rule",
            direction="ingress"
        )
        self._check_sg_rule_in_ovn_nb_db(new_rule['id'],
                                         neutron.STATEFUL_OVN_ACTION)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from datetime import datetime
import math
import random
import re
import socket
import time
import urllib.parse

import netaddr
from oslo_log import log

import tobiko
from tobiko.openstack import glance
from tobiko.openstack import keystone
from tobiko.openstack import stacks
from tobiko.openstack import tests
from tobiko.openstack import topology
from tobiko.tests.faults.ha import test_cloud_recovery
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.tripleo import containers
from tobiko.tripleo import nova
from tobiko.tripleo import pacemaker
from tobiko.tripleo import topology as tripleo_topology


LOG = log.getLogger(__name__)

network_disruption = """
 sudo iptables-save > ~/working.iptables.rules &&
 sudo iptables -I INPUT 1 -m state --state RELATED,ESTABLISHED -j ACCEPT &&
 sudo iptables -I INPUT 2 -p tcp -m state --state NEW -m tcp --dport 22 -j \
 ACCEPT &&
 sudo iptables -I INPUT 3 ! -i lo -j DROP &&
 sudo iptables -I OUTPUT 1 -p tcp --sport 22 -j ACCEPT &&
 sudo iptables -I OUTPUT 2 ! -o lo -j DROP
"""

undisrupt_network = """
 sudo iptables-restore ~/working.iptables.rules
"""
ovn_db_pcs_resource_restart = "sudo pcs resource restart ovn-dbs-bundle"
kill_rabbit = "sudo pkill -9 beam.smp"
remove_grastate = "sudo rm -rf /var/lib/mysql/grastate.dat"
check_bootstrap = """ps -eo lstart,cmd | grep -v grep|
grep wsrep-cluster-address=gcomm://"""
disable_galera = "sudo pcs resource disable galera --wait=60 2>&1"
enable_galera = "sudo pcs resource enable galera --wait=90 2>&1"
disable_haproxy = "sudo pcs resource disable haproxy-bundle --wait=30 2>&1"
enable_haproxy = "sudo pcs resource enable haproxy-bundle --wait=60 2>&1"
galera_sst_request = """sudo grep 'wsrep_sst_rsync.*'
/var/log/containers/mysql/mysqld.log"""
kill_mysqld = "sudo pkill -9 mysqld"
kill_mariadbd = "sudo pkill -9 mariadbd"


class PcsDisableException(tobiko.TobikoException):
    message = "pcs disable didn't shut down the resource"


class PcsEnableException(tobiko.TobikoException):
    message = "pcs enable didn't start the resource"


class GaleraBoostrapException(tobiko.TobikoException):
    message = "Bootstrap should not be done from node without grastate.dat"


class TimestampException(tobiko.TobikoException):
    message = "Timestamp mismatch: sst was requested before grastate removal"


def network_disrupt_node(node_name, disrupt_method=network_disruption):
    disrupt_node(node_name, disrupt_method=disrupt_method)


def network_undisrupt_node(node_name, disrupt_method=undisrupt_network):
    disrupt_node(node_name, disrupt_method=disrupt_method)


def disrupt_node(node_name, disrupt_method=network_disruption):

    # reboot all controllers and wait for ssh Up on them
    # hard reset is simultaneous while soft is sequential
    # method : method of disruption to use : network_disruption |
    # container_restart

    # using ssh_client.connect we use a fire and forget reboot method
    node = tripleo_topology.get_node(node_name)
    node.ssh_client.connect().exec_command(disrupt_method)
    LOG.info('disrupt exec: {} on server: {}'.format(disrupt_method,
                                                     node.name))
    check_overcloud_node_responsive(node)


def reboot_node(node_name, wait=True, reboot_method=sh.hard_reset_method):

    # reboot a node and wait for ssh Up on them
    # hard reset is simultaneous while soft is sequential
    # method : method of disruption to use : reset | network_disruption

    # using ssh_client.connect we use a fire and forget reboot method
    node = tripleo_topology.get_node(node_name)
    sh.reboot_host(ssh_client=node.ssh_client, wait=wait, method=reboot_method)
    LOG.info('disrupt exec: {} on server: {}'.format(reboot_method,
                                                     node.name))


def check_overcloud_node_responsive(node):
    node_checked = sh.execute("hostname",
                              ssh_client=node.ssh_client,
                              expect_exit_status=None).stdout
    LOG.info('{} is up '.format(node_checked))

    tobiko.cleanup_fixture(node.ssh_client)


def network_disrupt_all_controller_nodes(disrupt_method=network_disruption,
                                         exclude_list=None):
    disrupt_all_controller_nodes(disrupt_method=disrupt_method,
                                 exclude_list=exclude_list)


def reset_all_controller_nodes(disrupt_method=sh.hard_reset_method,
                               exclude_list=None):
    disrupt_all_controller_nodes(disrupt_method=disrupt_method,
                                 exclude_list=exclude_list)


def reset_all_controller_nodes_sequentially(
        disrupt_method=sh.hard_reset_method,
        sequentially=True, exclude_list=None):
    disrupt_all_controller_nodes(disrupt_method=disrupt_method,
                                 sequentially=sequentially,
                                 exclude_list=exclude_list)


def disrupt_all_controller_nodes(disrupt_method=sh.hard_reset_method,
                                 sequentially=False, exclude_list=None):
    # TODO(eolivare): join disrupt_all_controller_nodes and
    # reboot_all_controller_nodes methods because they are very similar

    # reboot all controllers and wait for ssh Up on them
    # method : method of disruptino to use : reset | network_disruption
    # hard reset is simultaneous while soft is sequential
    # exclude_list = list of nodes to NOT reset

    controlplane_groups = ['controller', 'messaging', 'database', 'networker']
    actual_controlplane_groups = tripleo_topology.actual_node_groups(
        controlplane_groups)
    nodes = topology.list_openstack_nodes(group=actual_controlplane_groups)

    # remove excluded nodes from reset list
    if exclude_list:
        nodes = [node for node in nodes if node.name not in exclude_list]

    for controller in nodes:
        if isinstance(disrupt_method, sh.RebootHostMethod):
            reboot_node(controller.name, wait=sequentially,
                        reboot_method=disrupt_method)
        else:
            # using ssh_client.connect we use a fire and forget reboot method
            controller.ssh_client.connect().exec_command(disrupt_method)
            LOG.info('disrupt exec: {} on server: {}'.format(disrupt_method,
                                                             controller.name))
            tobiko.cleanup_fixture(controller.ssh_client)
            if sequentially:
                check_overcloud_node_responsive(controller)
    if not sequentially:
        for controller in nodes:
            check_overcloud_node_responsive(controller)


def reboot_all_controller_nodes(reboot_method=sh.hard_reset_method,
                                sequentially=False, exclude_list=None):
    # reboot all controllers and wait for ssh Up on them
    # method : method of disruptino to use : hard or soft reset
    # hard reset is simultaneous while soft is sequential
    # exclude_list = list of nodes to NOT reset

    controlplane_groups = ['controller', 'messaging', 'database', 'networker']
    actual_controlplane_groups = tripleo_topology.actual_node_groups(
        controlplane_groups)
    nodes = topology.list_openstack_nodes(group=actual_controlplane_groups)

    # remove excluded nodes from reset list
    if exclude_list:
        nodes = [node for node in nodes if node.name not in exclude_list]

    for controller in nodes:
        sh.reboot_host(ssh_client=controller.ssh_client, wait=sequentially,
                       method=reboot_method)
        LOG.info('reboot exec: {} on server: {}'.format(reboot_method,
                                                        controller.name))
        tobiko.cleanup_fixture(controller.ssh_client)
    if not sequentially:
        for controller in nodes:
            check_overcloud_node_responsive(controller)


def get_main_vip():
    """return the ip of the overcloud main vip.
    Retreive an ip address (ipv4/ipv6) from the auth_url."""
    auth_url = keystone.default_keystone_credentials().auth_url
    auth_url_parsed = urllib.parse.urlsplit(auth_url)
    main_vip = auth_url_parsed.hostname

    if not (netaddr.valid_ipv4(main_vip) or netaddr.valid_ipv6(main_vip)):
        try:
            # socket.gethostbyname translates hostname to IPv4 - it fails when
            # no IPv4 address is available
            main_vip = socket.gethostbyname(main_vip)
        except socket.gaierror:
            # the following method obtains an IPv6 from a hostname
            main_vip = socket.getaddrinfo(
                main_vip, None, socket.AF_INET6)[0][4][0]

    return main_vip


def get_main_vip_controller(main_vip):
    """return the controller hostname ,
    which is holding the main_vip pacemaker resource"""
    # when the main_vip is ipv6, the pacemaker command output replaces : by .
    # we need to adapt the value accordingly
    main_vim_controller = pacemaker.get_overcloud_nodes_running_pcs_resource(
        resource=f"ip-{main_vip.replace(':', '.')}")[0]
    return main_vim_controller


def delete_evacuable_tagged_image():
    # delete evacuable tagged image because it prevents
    # non tagged evacuations if exists
    for img in glance.list_images():
        if 'evacuable' in img['tags']:
            glance.delete_image(img.id)


def disrupt_controller_main_vip(disrupt_method=sh.hard_reset_method,
                                inverse=False):

    # reset the controller holding the main vip (os_auth_url)
    # ip resource (managed via pacemaker)
    # find main vip by getting it from
    main_vip = get_main_vip()

    # find the node holding that resource via :

    main_vip_controller = get_main_vip_controller(main_vip)

    if isinstance(disrupt_method, sh.RebootHostMethod):
        if inverse:
            reboot_all_controller_nodes(reboot_method=disrupt_method,
                                        exclude_list=[main_vip_controller])
        else:
            reboot_node(main_vip_controller, reboot_method=disrupt_method)
    else:
        if inverse:
            # inverse the nodes reset selection
            disrupt_all_controller_nodes(disrupt_method=disrupt_method,
                                         exclude_list=[main_vip_controller])
        else:
            # get that node's ssh_client and reset it
            disrupt_node(main_vip_controller, disrupt_method=disrupt_method)


def reset_controller_main_vip():
    disrupt_controller_main_vip(disrupt_method=sh.hard_reset_method)


def reset_controllers_non_main_vip():
    disrupt_controller_main_vip(disrupt_method=sh.hard_reset_method,
                                inverse=True)


def crash_controller_main_vip():
    disrupt_controller_main_vip(disrupt_method=sh.crash_method)


def crash_controllers_non_main_vip():
    disrupt_controller_main_vip(disrupt_method=sh.crash_method,
                                inverse=True)


def network_disrupt_controller_main_vip():
    disrupt_controller_main_vip(disrupt_method=network_disruption)
    LOG.info('waiting 60s to avoid race conditions...')
    time.sleep(60.0)


def network_undisrupt_controller_main_vip():
    disrupt_controller_main_vip(disrupt_method=undisrupt_network)


def network_disrupt_controllers_non_main_vip():
    disrupt_controller_main_vip(disrupt_method=network_disruption,
                                inverse=True)


def network_undisrupt_controllers_non_main_vip():
    disrupt_controller_main_vip(disrupt_method=undisrupt_network,
                                inverse=True)


def reset_all_compute_nodes(hard_reset=False):

    # reboot all computes and wait for ssh Up on them
    # hard reset is simultaneous while soft is sequential
    if hard_reset:
        reset_method = sh.hard_reset_method
    else:
        reset_method = sh.soft_reset_method
    for compute in topology.list_openstack_nodes(group='compute'):
        # using ssh_client.connect we use a fire and forget reboot method
        sh.reboot_host(ssh_client=compute.ssh_client, wait=False,
                       method=reset_method)
        LOG.info('reboot exec:  {} on server: {}'.format(reset_method,
                                                         compute.name))
        tobiko.cleanup_fixture(compute.ssh_client)

    for compute in topology.list_openstack_nodes(group='compute'):
        compute_checked = sh.execute("hostname", ssh_client=compute.ssh_client,
                                     expect_exit_status=None).stdout
        LOG.info('{} is up '.format(compute_checked))


def reset_ovndb_pcs_master_resource():
    """restart ovndb pacemaker resource
    this method only restart the resource running on the controller with is
    acting as Master"""
    node = pacemaker.get_overcloud_nodes_running_pcs_resource(
        resource_type='(ocf::ovn:ovndb-servers):', resource_state='Master')[0]
    ovn_db_pcs_master_resource_restart = (ovn_db_pcs_resource_restart + ' ' +
                                          node)
    disrupt_node(node, disrupt_method=ovn_db_pcs_master_resource_restart)


def reset_ovndb_pcs_resource():
    """restart ovndb pacemaker resource
    this method restart the whole resource, i.e. on all the controller nodes"""
    node = pacemaker.get_overcloud_nodes_running_pcs_resource(
        resource_type='(ocf::ovn:ovndb-servers):', resource_state='Master')[0]
    disrupt_node(node, disrupt_method=ovn_db_pcs_resource_restart)


def reset_ovndb_master_container():
    """get and restart the ovndb master container
    use of partial name :  resource: ovn-dbs-bundle-0 =>
    container: ovn-dbs-bundle-podman-0 or ovn-dbs-bundle-docker-0"""
    node = pacemaker.get_overcloud_nodes_running_pcs_resource(
        resource_type='(ocf::ovn:ovndb-servers):', resource_state='Master')[0]
    resource = pacemaker.get_overcloud_resource(
        resource_type='(ocf::ovn:ovndb-servers):', resource_state='Master')
    resource = resource[0].rsplit('-', 1)[0]
    containers.action_on_container('restart',
                                   partial_container_name=resource,
                                   container_host=node)


def kill_rabbitmq_service():
    """kill a rabbit process on a random controller,
    check in pacemaker it is down"""
    if tripleo_topology.is_composable_roles_env():
        nodes = topology.list_openstack_nodes(group='messaging')
    else:
        nodes = topology.list_openstack_nodes(group='controller')
    node = random.choice(nodes)
    sh.execute(kill_rabbit, ssh_client=node.ssh_client)
    LOG.info('kill rabbit: {} on server: {}'.format(kill_rabbit,
             node.name))
    retry = tobiko.retry(timeout=30, interval=5)
    for _ in retry:
        if not(pacemaker.PacemakerResourcesStatus().
               rabbitmq_resource_healthy()):
            return


def kill_all_galera_services():
    """kill all galera processes,
    check in pacemaker it is down"""
    if tripleo_topology.is_composable_roles_env():
        nodes = topology.list_openstack_nodes(group='database')
    else:
        nodes = topology.list_openstack_nodes(group='controller')
    for node in nodes:
        if topology.verify_osp_version('17.0', lower=True):
            sh.execute(kill_mysqld, ssh_client=node.ssh_client)
            LOG.info('kill galera: {} on server: {}'.format(kill_mysqld,
                                                            node.name))
        else:
            sh.execute(kill_mariadbd, ssh_client=node.ssh_client)
            LOG.info('kill galera: {} on server: {}'.format(kill_mariadbd,
                                                            node.name))
    retry = tobiko.retry(timeout=30, interval=5)
    for _ in retry:
        if not(pacemaker.PacemakerResourcesStatus().
               galera_resource_healthy()):
            return


def remove_all_grastate_galera():
    """shut down galera properly,
    remove all grastate"""
    if tripleo_topology.is_composable_roles_env():
        nodes = topology.list_openstack_nodes(group='database')
    else:
        nodes = topology.list_openstack_nodes(group='controller')
    LOG.info('shut down galera: {} on all servers: {}'.
             format(disable_galera, nodes))
    if "resource 'galera' is not running on any node" not in\
            sh.execute(disable_galera, ssh_client=nodes[0].ssh_client).stdout:
        raise PcsDisableException()
    for node in nodes:
        sh.execute(remove_grastate, ssh_client=node.ssh_client)

    LOG.info('enable back galera: {} on all servers: {}'.
             format(enable_galera, nodes))
    if topology.verify_osp_version('17.0', lower=True):
        promoted = "master"
    else:
        promoted = "promoted"
    if f"resource 'galera' is {promoted} on node" not in\
            sh.execute(enable_galera, ssh_client=nodes[0].ssh_client).stdout:
        raise PcsEnableException()


def remove_one_grastate_galera():
    """shut down galera properly,
    delete /var/lib/mysql/grastate.dat in a random node,
    check that bootstrap is done from a node with grastate"""
    if tripleo_topology.is_composable_roles_env():
        nodes = topology.list_openstack_nodes(group='database')
    else:
        nodes = topology.list_openstack_nodes(group='controller')
    node = random.choice(nodes)
    LOG.info('disable haproxy-bunble')
    if "resource 'haproxy-bundle' is not running on any node" not in\
            sh.execute(disable_haproxy, ssh_client=node.ssh_client).stdout:
        raise PcsDisableException()
    LOG.info('shut down galera: {} on all servers: {}'.
             format(disable_galera, nodes))
    if "resource 'galera' is not running on any node" not in\
            sh.execute(disable_galera, ssh_client=node.ssh_client).stdout:
        raise PcsDisableException()
    LOG.info('remove grastate: {} on server: {}'.format(remove_grastate,
                                                        node.name))
    sh.execute(remove_grastate, ssh_client=node.ssh_client)

    LOG.info('enable back galera: {} on all servers: {}'.
             format(enable_galera, nodes))
    if topology.verify_osp_version('17.0', lower=True):
        promoted = "master"
    else:
        promoted = "promoted"
    if f"resource 'galera' is {promoted} on node" not in\
            sh.execute(enable_galera, ssh_client=node.ssh_client).stdout:
        raise PcsEnableException()
    LOG.info('enable haproxy-bundle')
    if "resource 'haproxy-bundle' is running on node" not in\
            sh.execute(enable_haproxy, ssh_client=node.ssh_client).stdout:
        raise PcsEnableException()
    # gcomm:// without args means that bootstrap is done from this node
    bootstrap = sh.execute(check_bootstrap, ssh_client=node.ssh_client).stdout
    if re.search('wsrep-cluster-address=gcomm:// --', bootstrap) is not None:
        raise GaleraBoostrapException()
    lastDate = re.findall(r"\w{,3}\s*\w{,3}\s*\d{,2}\s*\d{,2}:\d{,2}:\d{,2}\s*"
                          r"\d{4}", bootstrap)[-1]
    return node, lastDate


def request_galera_sst():
    """remove_one_grastate_galera,
    check that sst is requested by a node with grastate"""
    node, date = remove_one_grastate_galera()
    bootstrapDate = datetime.strptime(date, '%a %b %d %H:%M:%S %Y')
    retry = tobiko.retry(timeout=30, interval=5)
    for _ in retry:
        sst_req = sh.execute(galera_sst_request,
                             ssh_client=node.ssh_client).stdout
        if sst_req:
            break
    sstDate = datetime.strptime(re.findall
                                (r"\d{4}-\d{,2}-\d{,2}\s*\d{,2}:\d{,2}:\d{,2}",
                                 sst_req)[-1], '%Y-%m-%d %H:%M:%S')
    if bootstrapDate > sstDate:
        raise TimestampException


def evac_failover_compute(compute_host, failover_type=sh.hard_reset_method):
    """disrupt a compute, to trigger it's instance-HA evacuation
    failover_type=hard_reset_method etc.."""
    if failover_type in (sh.hard_reset_method, sh.soft_reset_method):
        reboot_node(compute_host, reboot_method=failover_type)
    else:
        disrupt_node(compute_host, disrupt_method=failover_type)


def check_iha_evacuation(failover_type=None, vm_type=None):
    """check vms on compute host,disrupt compute host,
    check all vms evacuated and pingable"""
    for iteration in range(2):
        LOG.info(f'Begin IHA tests iteration {iteration}')
        LOG.info('create 2 vms')
        tests.test_servers_creation(number_of_servers=2)
        compute_host = nova.get_random_compute_with_vms_name()
        vms_starting_state_df = nova.get_compute_vms_df(compute_host)
        if vm_type == 'shutoff':
            nova.stop_all_instances()
        if vm_type == 'evac_image_vm':
            evac_vm_stack = tests.test_evacuable_server_creation()
            evac_vm_id = nova.get_stack_server_id(evac_vm_stack)
            org_nova_evac_df = nova.vm_df(evac_vm_id, nova.get_vms_table())
        if not vm_type == 'shutoff':
            nova.check_df_vms_ping(vms_starting_state_df)
        LOG.info(f'perform a failover on {compute_host}')
        evac_failover_compute(compute_host, failover_type=failover_type)
        test_cloud_recovery.overcloud_health_checks(passive_checks_only=True)
        if vm_type == 'evac_image_vm':
            nova.check_vm_evacuations(vms_df_old=org_nova_evac_df,
                                      compute_host=compute_host,
                                      timeout=600,
                                      check_no_evacuation=True)
            # delete evacuable tagged image because it prevents
            # non tagged evacuations if exists
            delete_evacuable_tagged_image()
            new_nova_evac_df = nova.vm_df(evac_vm_id, nova.get_vms_table())
            nova.check_vm_evacuations(org_nova_evac_df, new_nova_evac_df)
        else:
            nova.check_vm_evacuations(vms_df_old=vms_starting_state_df,
                                      compute_host=compute_host,
                                      timeout=600)
        LOG.info('check evac is Done')
        if not vm_type == 'shutoff':
            nova.check_df_vms_ping(vms_starting_state_df)


def check_iha_evacuation_evac_image_vm():
    check_iha_evacuation(failover_type=sh.hard_reset_method,
                         vm_type='evac_image_vm')


def check_iha_evacuation_hard_reset():
    check_iha_evacuation(failover_type=sh.hard_reset_method)


def check_iha_evacuation_network_disruption():
    check_iha_evacuation(failover_type=network_disruption)


def check_iha_evacuation_hard_reset_shutoff_instance():
    check_iha_evacuation(failover_type=sh.hard_reset_method, vm_type='shutoff')


def test_controllers_shutdown():
    test_case = tobiko.get_test_case()

    all_nodes = topology.list_openstack_nodes(group='controller')
    if len(all_nodes) < 3:
        tobiko.skip_test('It requires at least three controller nodes')

    all_node_names = [node.name for node in all_nodes]
    LOG.info("Ensure all controller nodes are running: "
             f"{all_node_names}")
    for node in all_nodes:
        node.power_on_overcloud_node()
    topology.assert_reachable_nodes(all_nodes)

    LOG.debug('Check VM is running while all controllers nodes are on')
    nova_server = tobiko.setup_fixture(stacks.CirrosServerStackFixture)
    nova_server_ip = nova_server.ip_address
    ping.assert_reachable_hosts([nova_server_ip])

    quorum_level = math.ceil(0.5 * len(all_nodes))
    assert quorum_level >= len(all_nodes) - quorum_level
    nodes = random.sample(all_nodes, quorum_level)
    node_names = [node.name for node in nodes]
    LOG.info(f"Power off {quorum_level} random controller nodes: "
             f"{node_names}")
    for node in nodes:
        node.power_off_overcloud_node()
        test_case.addCleanup(node.power_on_overcloud_node)
    topology.assert_unreachable_nodes(nodes, retry_count=1)
    topology.assert_reachable_nodes(node
                                    for node in all_nodes
                                    if node not in nodes)

    LOG.debug('Check whenever VM is still running while some "'
              '"controllers nodes are off')
    reachable, unreachable = ping.ping_hosts([nova_server_ip],
                                             count=1)
    if reachable:
        LOG.debug(f"VM ips are reachable: {reachable}")
    if unreachable:
        LOG.debug(f"VM is are unreachable: {unreachable}")
    # TODO what do we expect here: VM reachable or unreachable?

    random.shuffle(nodes)
    LOG.info(f"Power on controller nodes: {node_names}")
    for node in nodes:
        node.power_on_overcloud_node()

    LOG.debug("Check all controller nodes are running again: "
              f"{all_node_names}")
    topology.assert_reachable_nodes(all_nodes, retry_timeout=600.)

    LOG.debug('Check VM is running while all controllers nodes are on')
    ping.assert_reachable_hosts([nova_server_ip])



from __future__ import absolute_import

import abc
import functools
import os
import re
import time
import typing

from oslo_log import log
import pandas

import tobiko
from tobiko import config
from tobiko import podman
from tobiko import docker
from tobiko.openstack import neutron
from tobiko.openstack import topology
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.tripleo import overcloud
from tobiko.tripleo import topology as tripleo_topology


CONF = config.CONF
LOG = log.getLogger(__name__)


class ContainerRuntime(abc.ABC):
    runtime_name: str
    version_pattern: typing.Pattern

    def match_version(self, version: str) -> bool:
        for version_line in version.splitlines():
            if self.version_pattern.match(version_line) is not None:
                return True
        return False

    def get_client(self, ssh_client):
        for attempt in tobiko.retry(timeout=60.0,
                                    interval=5.0):
            try:
                client = self._get_client(ssh_client=ssh_client)
                break
            # TODO chose a better exception type
            except Exception:
                if attempt.is_last:
                    raise
                LOG.debug('Unable to connect to docker server',
                          exc_info=1)
                ssh.reset_default_ssh_port_forward_manager()
        else:
            raise RuntimeError("Broken retry loop")
        return client

    def _get_client(self, ssh_client):
        raise NotImplementedError

    def list_containers(self, ssh_client):
        raise NotImplementedError


class DockerContainerRuntime(ContainerRuntime):
    runtime_name = 'docker'
    version_pattern = re.compile('Docker version .*', re.IGNORECASE)

    def _get_client(self, ssh_client):
        return docker.get_docker_client(ssh_client=ssh_client,
                                        sudo=True).connect()

    def list_containers(self, ssh_client):
        client = self.get_client(ssh_client=ssh_client)
        return docker.list_docker_containers(client=client)


class PodmanContainerRuntime(ContainerRuntime):
    runtime_name = 'podman'
    version_pattern = re.compile('Podman version .*', re.IGNORECASE)

    def _get_client(self, ssh_client):
        return podman.get_podman_client(ssh_client=ssh_client).connect()

    def list_containers(self, ssh_client):
        client = self.get_client(ssh_client=ssh_client)
        return podman.list_podman_containers(client=client)


DOCKER_RUNTIME = DockerContainerRuntime()
PODMAN_RUNTIME = PodmanContainerRuntime()
CONTAINER_RUNTIMES = [PODMAN_RUNTIME, DOCKER_RUNTIME]


class ContainerRuntimeFixture(tobiko.SharedFixture):

    runtime: typing.Optional[ContainerRuntime] = None

    def setup_fixture(self):
        if overcloud.has_overcloud():
            self.runtime = self.get_runtime()

    def cleanup_fixture(self):
        self.runtime = None

    @staticmethod
    def get_runtime() -> typing.Optional[ContainerRuntime]:
        """check what container runtime is running
        and return a handle to it"""
        # TODO THIS LOCKS SSH CLIENT TO CONTROLLER
        for node in topology.list_openstack_nodes(group='controller'):
            try:
                result = sh.execute('podman --version || docker --version',
                                    ssh_client=node.ssh_client)
            except sh.ShellCommandFailed:
                continue
            for runtime in CONTAINER_RUNTIMES:
                for version in [result.stdout, result.stderr]:
                    if runtime.match_version(version):
                        return runtime
        raise RuntimeError(
            "Unable to find any container runtime in any overcloud "
            "controller node")


def get_container_runtime() -> ContainerRuntime:
    runtime = tobiko.setup_fixture(ContainerRuntimeFixture).runtime
    return runtime


def get_container_runtime_name() -> str:
    return get_container_runtime().runtime_name


def is_docker() -> bool:
    return get_container_runtime().runtime_name == 'docker'


def is_podman() -> bool:
    return get_container_runtime().runtime_name == 'podman'


def has_container_runtime() -> bool:
    return get_container_runtime() is not None


def skip_unless_has_container_runtime():
    return tobiko.skip_unless('Container runtime not found',
                              has_container_runtime)


@functools.lru_cache()
def list_node_containers(ssh_client):
    """returns a list of containers and their run state"""
    return get_container_runtime().list_containers(ssh_client=ssh_client)


def get_container_client(ssh_client=None):
    """returns a list of containers and their run state"""
    return get_container_runtime().get_client(ssh_client=ssh_client)


def list_containers_df(group=None):
    actual_containers_list = list_containers(group)
    return pandas.DataFrame(
        get_container_states_list(actual_containers_list),
        columns=['container_host', 'container_name', 'container_state'])


def list_containers(group=None):
    """get list of containers in running state
    from specified node group
    returns : a list of overcloud_node's running containers"""

    # moved here from topology
    # reason : Workaround for :
    # AttributeError: module 'tobiko.openstack.topology' has no
    # attribute 'container_runtime'

    if group is None:
        group = 'overcloud'
    containers_list = tobiko.Selection()
    openstack_nodes = topology.list_openstack_nodes(group=group)

    for node in openstack_nodes:
        LOG.debug(f"List containers for node {node.name}")
        node_containers_list = list_node_containers(ssh_client=node.ssh_client)
        containers_list.extend(node_containers_list)
    return containers_list


expected_containers_file = os.path.expanduser(
    '~/expected_containers_list_df.csv')


def save_containers_state_to_file(expected_containers_list,):
    expected_containers_list_df = pandas.DataFrame(
        get_container_states_list(expected_containers_list),
        columns=['container_host', 'container_name', 'container_state'])
    expected_containers_list_df.to_csv(
        expected_containers_file)
    return expected_containers_file


class ContainerMismatchException(tobiko.TobikoException):
    pass


def assert_containers_running(group, expected_containers, full_name=True,
                              bool_check=False):

    """assert that all containers specified in the list are running
    on the specified openstack group(controller or compute etc..)
    if bool_check is True then return only True or false without failing"""

    if is_docker():
        LOG.info('not checking common containers since we are on docker')
        return

    failures = []

    openstack_nodes = topology.list_openstack_nodes(group=group)
    for node in openstack_nodes:
        node_containers = list_node_containers(ssh_client=node.ssh_client)
        containers_list_df = pandas.DataFrame(
            get_container_states_list(node_containers),
            columns=['container_host', 'container_name', 'container_state'])
        # check that the containers are present
        LOG.info('node: {} containers list : {}'.format(
            node.name, containers_list_df.to_string(index=False)))
        for container in expected_containers:
            # get container attrs dataframe
            if full_name:
                container_attrs = containers_list_df.query(
                    'container_name == "{}"'.format(container))
            else:
                container_attrs = containers_list_df[
                    containers_list_df['container_name'].
                    str.contains(container)]
            # check if the container exists
            LOG.info('checking container: {}'.format(container))
            if container_attrs.empty:
                failures.append(
                    'expected container {} not found on node {} ! : \n\n'.
                    format(container, node.name))
            # if container exists, check it is running
            else:
                # only one running container is expected
                container_running_attrs = container_attrs.query(
                    'container_state=="running"')
                if container_running_attrs.empty:
                    failures.append(
                        'expected container {} is not running on node {} , '
                        'its state is {}! : \n\n'.format(
                            container, node.name,
                            container_attrs.container_state.values.item()))
                elif len(container_running_attrs) > 1:
                    failures.append(
                        'only one running container {} was expected on '
                        'node {}, but got {}! : \n\n'.format(
                            container, node.name,
                            len(container_running_attrs)))

    if not bool_check and failures:
        tobiko.fail(
            'container states mismatched:\n{}'.format('\n'.join(failures)),
            ContainerMismatchException)

    elif bool_check and failures:
        return False

    else:
        LOG.info('All specified containers are in running state! ')
        return True


def get_libvirt_container_name():
    if topology.verify_osp_version('17.0', lower=True):
        return 'nova_libvirt'
    else:
        return 'nova_virtqemud'


def assert_all_tripleo_containers_running():
    """check that all common tripleo containers are running
    param: group controller or compute , check containers
    sets in computes or controllers"""
    nova_libvirt = get_libvirt_container_name()

    common_controller_tripleo_containers = ['cinder_api', 'cinder_api_cron',
                                            'cinder_scheduler',
                                            'glance_api', 'heat_api',
                                            'heat_api_cfn',
                                            'heat_api_cron', 'heat_engine',
                                            'horizon', 'iscsid', 'keystone',
                                            'logrotate_crond', 'memcached',
                                            'neutron_api', 'nova_api',
                                            'nova_api_cron', 'nova_conductor',
                                            'nova_metadata', 'nova_scheduler',
                                            'nova_vnc_proxy',
                                            'swift_account_auditor',
                                            'swift_account_reaper',
                                            'swift_account_replicator',
                                            'swift_account_server',
                                            'swift_container_auditor',
                                            'swift_container_replicator',
                                            'swift_container_server',
                                            'swift_container_updater',
                                            'swift_object_auditor',
                                            'swift_object_expirer',
                                            'swift_object_replicator',
                                            'swift_object_server',
                                            'swift_object_updater',
                                            'swift_proxy', 'swift_rsync']

    shiftonstack_controller_tripleo_containers = ['cinder_api',
                                                  'cinder_api_cron',
                                                  'cinder_scheduler',
                                                  'glance_api', 'heat_api',
                                                  'heat_api_cfn',
                                                  'heat_api_cron',
                                                  'heat_engine',
                                                  'horizon', 'iscsid',
                                                  'keystone',
                                                  'logrotate_crond',
                                                  'memcached',
                                                  'neutron_api', 'nova_api',
                                                  'nova_api_cron',
                                                  'nova_conductor',
                                                  'nova_metadata',
                                                  'nova_scheduler',
                                                  'nova_vnc_proxy']

    common_compute_tripleo_containers = ['iscsid', 'logrotate_crond',
                                         'nova_compute', nova_libvirt,
                                         'nova_migration_target',
                                         'nova_virtlogd']

    if ceph_rgw_expected():
        tripleo_containers_to_check = \
            shiftonstack_controller_tripleo_containers
    else:
        tripleo_containers_to_check = \
            common_controller_tripleo_containers

    for group, group_containers in [('controller',
                                     tripleo_containers_to_check),
                                    ('compute',
                                     common_compute_tripleo_containers)]:
        assert_containers_running(group, group_containers)
    # TODO: need to address OSP-version specific containers here.
    # optional ovn containers checks
    assert_ovn_containers_running()


def assert_ceph_rgw_container_running():
    """check if ceph-rgw is running, usually on
    shifton stack deployments , if so we should not check for
    swift containers precence"""
    return assert_containers_running('controller', ['ceph-rgw'],
                                     full_name=False, bool_check=True)


def ceph_rgw_expected():
    """Returns True if 'ceph_rgw' parameter in the config file is set to 'true'
    or ceph rgw containers were detected running by Tobiko.
    """
    return CONF.tobiko.tripleo.ceph_rgw or assert_ceph_rgw_container_running()


def skip_if_ceph_rgw():
    return tobiko.skip_if('Ceph rgw deployed', ceph_rgw_expected)


def osp13_container_name_short_format(container_name_long_format):
    """This takes a long format container name :
    'rhosp13/openstack-neutron-l3-agent'
    and turns it into : neutron_l3_agent
    """
    return re.sub('-', '_', re.sub('rhosp13/openstack-', '',
                                   container_name_long_format))


def assert_ovn_containers_running():
    if not neutron.has_ovn():
        LOG.info("Networking OVN not configured")
        return

    # specific OVN verifications
    ovn_controller_containers = ['ovn_controller']
    if overcloud.is_ovn_using_raft():
        ovn_controller_containers += ['ovn_cluster_northd',
                                      'ovn_cluster_north_db_server',
                                      'ovn_cluster_south_db_server']
    else:
        container_runtime_name = get_container_runtime_name()
        ovn_controller_containers += [
            f'ovn-dbs-bundle-{container_runtime_name}-']
    ovn_compute_containers = ['ovn_metadata_agent',
                              'ovn_controller']
    group_containers_list = [('controller', ovn_controller_containers),
                             ('compute', ovn_compute_containers)]
    if 'networker' in topology.list_openstack_node_groups():
        ovn_networker_containers = ['ovn_controller']
        group_containers_list.append(('networker',
                                      ovn_networker_containers))
    for group, group_containers in group_containers_list:
        assert_containers_running(group, group_containers, full_name=False)
    LOG.info("Networking OVN containers verified in running state")


def run_container_config_validations():
    """check containers configuration in different scenarios
    """

    # TODO add here any generic configuration validation
    config_checkings = []

    if neutron.has_ovn():
        ovn_config_checkings = \
            [{'node_group': 'controller',
              'container_name': 'neutron_api',
              'config_file': '/etc/neutron/plugins/ml2/ml2_conf.ini',
              'param_validations': [{'section': 'ml2',
                                     'param': 'mechanism_drivers',
                                     'expected_value': 'ovn'},
                                    {'section': 'ml2',
                                     'param': 'type_drivers',
                                     'expected_value': 'geneve'},
                                    {'section': 'ovn',
                                     'param': 'ovn_metadata_enabled',
                                     'expected_value': 'True'}]}]
        config_checkings += ovn_config_checkings
    else:
        ovs_config_checkings = \
            [{'node_group': 'controller',
              'container_name': 'neutron_api',
              'config_file': '/etc/neutron/plugins/ml2/ml2_conf.ini',
              'param_validations': [{'section': 'ml2',
                                     'param': 'mechanism_drivers',
                                     'expected_value': 'openvswitch'}]}]
        config_checkings += ovs_config_checkings

    container_runtime_name = get_container_runtime_name()
    for config_check in config_checkings:
        for node in topology.list_openstack_nodes(
                group=config_check['node_group']):
            for param_check in config_check['param_validations']:
                obtained_param = sh.execute(
                    f"{container_runtime_name} exec -uroot "
                    f"{config_check['container_name']} crudini "
                    f"--get {config_check['config_file']} "
                    f"{param_check['section']} {param_check['param']}",
                    ssh_client=node.ssh_client, sudo=True).stdout.strip()
                if param_check['expected_value'] not in obtained_param:
                    tobiko.fail(f"Expected {param_check['param']} value: "
                                f"{param_check['expected_value']}\n"
                                f"Obtained {param_check['param']} value: "
                                f"{obtained_param}")
        LOG.info("Configuration verified:\n"
                 f"node group: {config_check['node_group']}\n"
                 f"container: {config_check['container_name']}\n"
                 f"config file: {config_check['config_file']}")


def comparable_container_keys(container, include_container_objects=False):
    """returns the tuple : 'container_host','container_name',
    'container_state, container object if specified'
     """
    # Differenciate between podman_ver3 with podman-py from earlier api
    if is_podman():
        if podman.Podman_Version_3():
            con_host_name_stat_obj_tuple = (tripleo_topology.ip_to_hostname(
                container.client.base_url.netloc.rsplit('_')[1]),
                                            container.attrs[
                'Names'][0], container.attrs['State'], container)

            con_host_name_stat_tuple = (tripleo_topology.ip_to_hostname(
                container.client.base_url.netloc.rsplit('_')[1]),
                                        container.attrs[
                'Names'][0], container.attrs['State'])
        else:

            con_host_name_stat_obj_tuple = (tripleo_topology.ip_to_hostname(
                container._client._context.hostname),  # pylint: disable=W0212
                container.data['names'], container.data['status'], container)

            con_host_name_stat_tuple = (tripleo_topology.ip_to_hostname(
                container._client._context.hostname),  # pylint: disable=W0212
                container.data['names'], container.data['status'])

        if include_container_objects:
            return con_host_name_stat_obj_tuple
        else:
            return con_host_name_stat_tuple

    elif is_docker():
        if include_container_objects:
            return (container.client.api.ssh_client.hostname,
                    osp13_container_name_short_format(container.attrs[
                                                          'Labels']['name']),
                    container.attrs['State'],
                    container)
        else:
            return (container.client.api.ssh_client.hostname,
                    osp13_container_name_short_format(container.attrs[
                                                          'Labels']['name']),
                    container.attrs['State'])


@functools.lru_cache()
def list_containers_objects_df():
    containers_list = list_containers()
    containers_objects_list_df = pandas.DataFrame(
        get_container_states_list(
            containers_list, include_container_objects=True),
        columns=['container_host', 'container_name',
                 'container_state', 'container_object'])
    return containers_objects_list_df


def get_overcloud_container(container_name=None, container_host=None,
                            partial_container_name=None):
    """gets an container object by name on specified host
    container"""
    con_obj_df = list_containers_objects_df()
    if partial_container_name and container_host:
        con_obj_df = con_obj_df[con_obj_df['container_name'].str.contains(
            partial_container_name)]
        contaniner_obj = con_obj_df.query(
            'container_host == "{container_host}"'.format(
                container_host=container_host))['container_object']
    elif container_host:
        contaniner_obj = con_obj_df.query(
            'container_name == "{container_name}"'
            ' and container_host == "{container_host}"'.
            format(container_host=container_host,
                   container_name=container_name)).container_object
    else:
        contaniner_obj = con_obj_df.query(
            'container_name == "{container_name}"'.
            format(container_name=container_name)).container_object
    if not contaniner_obj.empty:
        return contaniner_obj.values[0]
    else:
        tobiko.fail('container {} not found!'.format(container_name))


def action_on_container(action: str,
                        container_name=None,
                        container_host=None,
                        partial_container_name=None):
    """take a container snd preform an action on it
    actions are as defined in : podman/libs/containers.py:14/164"""

    LOG.debug(f"Executing '{action}' action on container "
              f"'{container_name}@{container_host}'...")
    container = get_overcloud_container(
        container_name=container_name,
        container_host=container_host,
        partial_container_name=partial_container_name)

    container_class: typing.Type = type(container)
    # we get the specified action as function from podman lib
    action_method: typing.Optional[typing.Callable] = getattr(
        container_class, action, None)
    if action_method is None:
        raise TypeError(f"Unsupported container action for class :"
                        f" {container_class}")
    if not callable(action_method):
        raise TypeError(
            f"Attribute '{container_class.__qualname__}.{action}' value "
            f" is not a method: {action_method!r}")
    LOG.debug(f"Calling '{action_method}' action on container "
              f"'{container}'")
    return action_method(container)


def get_container_states_list(containers_list,
                              include_container_objects=False):
    container_states_list = tobiko.Selection()
    container_states_list.extend([comparable_container_keys(
        container, include_container_objects=include_container_objects) for
                                  container in containers_list])
    return container_states_list


pcs_resource_list = ['haproxy', 'galera', 'redis', 'ovn-dbs', 'cinder',
                     'rabbitmq', 'manila', 'ceph', 'pacemaker']


def remove_containers_if_pacemaker_resources(comparable_containers_df):
    """remove any containers in
    param: comparable_containers_df that are pacemaker resources
    i.e if they contain tha names of resources defined in
    pcs_resources_list"""

    for row in comparable_containers_df.iterrows():
        for pcs_resource in pcs_resource_list:
            if pcs_resource in str(row):
                LOG.info(f'pcs resource {pcs_resource} has changed state, '
                         f'but that\'s ok since pcs resources can change '
                         f'state/host: {str(row)}')
                # if a pcs resource is found , we drop that row
                comparable_containers_df.drop(row[0], inplace=True)
    return comparable_containers_df


def dataframe_difference(df1, df2, which=None):
    """Find rows which are different between two DataFrames."""
    comparison_df = df1.merge(df2,
                              indicator='same_state',
                              how='outer')
    # return only non identical rows
    if which is None:
        diff_df = comparison_df[comparison_df['same_state'] != 'both']

    else:
        diff_df = comparison_df[comparison_df['same_state'] == which]

    # if the list of diffrent state container are pacemaker resources ignore
    # the error since we are checking pacemaker also.

    remove_containers_if_pacemaker_resources(diff_df)

    return diff_df


def assert_equal_containers_state(expected_containers_list=None,
                                  timeout=120, interval=2,
                                  recreate_expected=False):

    """compare all overcloud container states with using two lists:
    one is current , the other some past list
    first time this method runs it creates a file holding overcloud
    containers' states: ~/expected_containers_list_df.csv'
    second time it creates a current containers states list and
    compares them, they must be identical"""

    # if we have a file or an explicit variable use that , otherwise  create
    # and return
    if recreate_expected or (not expected_containers_list and
                             not os.path.exists(expected_containers_file)):
        save_containers_state_to_file(list_containers())
        return

    elif expected_containers_list:
        expected_containers_list_df = pandas.DataFrame(
            get_container_states_list(expected_containers_list),
            columns=['container_host', 'container_name', 'container_state'])

    elif os.path.exists(expected_containers_file):
        expected_containers_list_df = pandas.read_csv(
            expected_containers_file)

    failures = []
    start = time.time()
    error_info = 'Output explanation: left_only is the original state, ' \
                 'right_only is the new state'

    while time.time() - start < timeout:

        failures = []
        actual_containers_list_df = list_containers_df()

        LOG.info('expected_containers_list_df: {} '.format(
            expected_containers_list_df.to_string(index=False)))
        LOG.info('actual_containers_list_df: {} '.format(
            actual_containers_list_df.to_string(index=False)))

        # execute a `dataframe` diff between the expected and actual containers
        expected_containers_state_changed = \
            dataframe_difference(expected_containers_list_df,
                                 actual_containers_list_df)
        # check for changed state containerstopology
        if not expected_containers_state_changed.empty:
            failures.append('expected containers changed state ! : '
                            '\n\n{}\n{}'.format(
                             expected_containers_state_changed.
                             to_string(index=False), error_info))
            LOG.info('container states mismatched:\n{}\n'.format(failures))
            time.sleep(interval)
            # clear cache to obtain new data
            list_node_containers.cache_clear()
        else:
            LOG.info("assert_equal_containers_state :"
                     " OK, all containers are on the same state")
            return
    if failures:
        tobiko.fail('container states mismatched:\n{!s}', '\n'.join(
            failures))



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import netaddr

import tobiko


class Iperf3ClientParameters(typing.NamedTuple):
    address: str
    bitrate: typing.Optional[int] = None
    download: typing.Optional[bool] = None
    port: typing.Optional[int] = None
    protocol: typing.Optional[str] = None
    timeout: typing.Optional[int] = None


def iperf3_client_parameters(
        address: typing.Union[str, netaddr.IPAddress],
        bitrate: int = None,
        download: bool = None,
        port: int = None,
        protocol: str = None,
        timeout: int = None):
    """Get iperf3 client parameters
    mode allowed values: client or server
    ip is only needed for client mode
    """
    config = tobiko.tobiko_config().iperf3
    if isinstance(address, netaddr.IPAddress):
        address = str(address)
    if bitrate is None:
        bitrate = config.bitrate
    if download is None:
        download = config.download
    if port is None:
        port = config.port
    if protocol is None:
        protocol = config.protocol
    if timeout is None:
        timeout = config.timeout
    return Iperf3ClientParameters(address=address,
                                  bitrate=bitrate,
                                  download=download,
                                  port=port,
                                  protocol=protocol,
                                  timeout=timeout)



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

from oslo_log import log
from validations_libs import validation_actions

from tobiko import tripleo
from tobiko.tripleo import overcloud


LOG = log.getLogger(__name__)


def run_post_deployment_validations():
    """run validtion framework post-deploument validations
    only if we're in a tripleo env"""

    if overcloud.has_overcloud():
        inventory_file = os.path.expanduser('~/hosts.yaml')
        tripleo.fetch_tripleo_inventary_file(inventory_file=inventory_file)
        failures = []
        validates_object = validation_actions.ValidationActions()
        try:
            validations_result = validates_object.run_validations(
                group='post-deployment',
                quiet=False,
                inventory=inventory_file)
        except Exception:
            LOG.exception('Validation lib unhandled errors')
            return

        for validation in validations_result:
            if validation['Status'] == 'FAILED':
                failures.append(
                    'failed validation: {}\n\n'.format(validation))
            elif validation['Status'] == 'PASSED':
                LOG.info('passed validation: {}\n\n'.format(validation))
        if failures:
            LOG.info('Failed tripleo validations:\n {}'.format(failures))
            # We should not fail over validations in the beginning we have to
            # test run them, and handle false negatives.

            # tobiko.fail(
            #         'nova agents are unhealthy:\n{!s}', '\n'.join(failures))

    # to list possible validations:
    # validates_object.list_validations()

    # single validation example
    # validates_object.run_validations(
    # validation_name='healthcheck-service-status',quiet=False,inventory='/home/stack/hosts.yaml')



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from keystoneauth1 import loading
from keystoneauth1 import session as _session
from keystoneauth1 import plugin as _plugin
from oslo_log import log

import tobiko
from tobiko.openstack.keystone import _credentials
from tobiko import http


LOG = log.getLogger(__name__)

KEYSTONE_SESSION_CLASSES = _session.Session,
KeystoneSession = typing.Union[_session.Session]


class KeystoneSessionFixture(tobiko.SharedFixture):

    VALID_CREDENTIALS_TYPES = (_credentials.KeystoneCredentials,
                               _credentials.KeystoneCredentialsFixture,
                               type)

    def __init__(self,
                 credentials: _credentials.KeystoneCredentialsType = None,
                 session: KeystoneSession = None):
        super(KeystoneSessionFixture, self).__init__()
        if credentials is not None:
            tobiko.check_valid_type(credentials, *self.VALID_CREDENTIALS_TYPES)
        self._credentials = credentials
        self.session = session

    @property
    def credentials(self) -> _credentials.KeystoneCredentials:
        if self._credentials is None:
            self._credentials = self._get_credentials()
        elif not isinstance(self._credentials,
                            _credentials.KeystoneCredentials):
            self._credentials = _credentials.keystone_credentials(
                self._credentials)
        return self._credentials

    def setup_fixture(self):
        self.setup_session()

    def setup_session(self):
        if self.session is None:
            self.session = self._get_session()

    def _get_session(self) -> KeystoneSession:
        credentials = self.credentials
        LOG.debug("Create Keystone session from credentials "
                  f"{credentials}")
        credentials.validate()
        loader = loading.get_plugin_loader('password')
        params = credentials.to_dict()
        # api version parameter is not accepted
        params.pop('api_version', None)
        params.pop('cacert', None)
        auth = loader.load_from_options(**params)
        session = _session.Session(auth=auth, verify=False)
        http.setup_http_session(session)
        return session

    @staticmethod
    def _get_credentials() -> _credentials.KeystoneCredentials:
        return _credentials.default_keystone_credentials()


KeystoneSessionType = typing.Union[KeystoneSession,
                                   KeystoneSessionFixture,
                                   typing.Type[KeystoneSessionFixture]]


def keystone_session(obj: KeystoneSessionType = None) -> KeystoneSession:
    if obj is None:
        return default_keystone_session()
    if tobiko.is_fixture(obj):
        obj = tobiko.get_fixture(obj)
        if isinstance(obj, KeystoneSessionFixture):
            obj = tobiko.setup_fixture(obj).session
    return tobiko.check_valid_type(obj, KEYSTONE_SESSION_CLASSES)


InitSessionType = typing.Callable[[_credentials.KeystoneCredentials],
                                  KeystoneSessionFixture]


class KeystoneSessionManager(object):

    def __init__(self):
        self.sessions: typing.Dict[_credentials.KeystoneCredentials,
                                   KeystoneSessionFixture] = {}

    def get_session(self,
                    credentials: _credentials.KeystoneCredentialsType = None,
                    init_session: InitSessionType = None,
                    shared: bool = True) \
            -> KeystoneSessionFixture:
        credentials = _credentials.keystone_credentials(credentials)
        if shared:
            session = self.sessions.get(credentials)
            if session is not None:
                return session
        session = self.create_session(credentials=credentials,
                                      init_session=init_session)
        if shared:
            self.sessions[credentials] = session
        return session

    def create_session(self,
                       credentials: _credentials.KeystoneCredentials,
                       init_session: InitSessionType = None) \
            -> KeystoneSessionFixture:
        if init_session is None:
            init_session = self.init_session
        assert callable(init_session)
        LOG.debug('Initialize Keystone session:\n'
                  f"  init_session: {init_session}\n"
                  f"  credentials: {credentials}\n")
        session = init_session(credentials)
        LOG.debug('Got new Keystone session:\n'
                  f"  init_session: {init_session}\n"
                  f"  credentials: {credentials}\n"
                  f"  session: {session}\n")
        return tobiko.check_valid_type(session, KeystoneSessionFixture)

    @staticmethod
    def init_session(credentials: _credentials.KeystoneCredentials) \
            -> KeystoneSessionFixture:
        return KeystoneSessionFixture(credentials=credentials)


KEYSTONE_SESSION_MANAGER = KeystoneSessionManager()


def default_keystone_session(
        shared: bool = True,
        init_session: InitSessionType = None,
        manager: KeystoneSessionManager = None) -> \
        KeystoneSession:
    return get_keystone_session(shared=shared,
                                init_session=init_session,
                                manager=manager)


def get_keystone_session(
        credentials: _credentials.KeystoneCredentialsType = None,
        shared: bool = True,
        init_session: typing.Any = None,
        manager: KeystoneSessionManager = None) -> \
        KeystoneSession:
    if manager is None:
        manager = KEYSTONE_SESSION_MANAGER
    session = manager.get_session(credentials=credentials,
                                  shared=shared,
                                  init_session=init_session)
    tobiko.check_valid_type(session, KeystoneSessionFixture)
    return tobiko.setup_fixture(session).session


def get_keystone_endpoint(
        session: KeystoneSessionType = None,
        auth: typing.Optional[_plugin.BaseAuthPlugin] = None,
        **kwargs) -> \
        typing.Optional[str]:
    return keystone_session(session).get_endpoint(auth=auth, **kwargs)


def get_keystone_token(
        session: KeystoneSessionType = None,
        auth: typing.Optional[_plugin.BaseAuthPlugin] = None) -> \
        typing.Optional[str]:
    return keystone_session(session).get_token(auth=auth)



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import testtools

from tobiko.openstack import neutron
from tobiko import tripleo
from tobiko.tripleo import pacemaker
from tobiko.tripleo import processes


LOG = log.getLogger(__name__)


@tripleo.skip_if_missing_overcloud
class PacemakerTest(testtools.TestCase):

    def test_resources_health(self):
        """Check cluster failed statuses"""
        self.assertTrue(pacemaker.PacemakerResourcesStatus().all_healthy)


@tripleo.skip_if_missing_overcloud
class OvercloudProcessesTest(testtools.TestCase):

    def test_overcloud_processes_running(self):
        procs = processes.OvercloudProcessesStatus()
        self.assertTrue(procs.basic_overcloud_processes_running)

    @neutron.skip_unless_is_ovn()
    def test_ovn_overcloud_processes_validations(self):
        procs = processes.OvercloudProcessesStatus()
        self.assertTrue(procs.ovn_overcloud_processes_validations)
